use threads;
use threads::shared;
use File::Basename;
my $dirname = dirname(__FILE__);
$SIG{INT} = \&interrupt;

##################### Settings ##################################

my $conserve_weight = 3;

#################################################################

##################### Filepaths #################################

my $ML_exe = "$dirname/maximum_likelihood_gamma_BLAS.exe";
my $BW_exe = "$dirname/baum_welch.exe";

#################################################################

$START = 1;
$STOP = 0;
$NO_LOCAL = 0;
$max_cutoff = 0;
$max_output = 0;
$posterior_threshold = 0.6;
$local_window = 20;
$cpu = 1;
$max_output = 0;
$r_length = 10;
$min_output = 1;
$flank = 0;
$debug = 0;
$SMOOTH = 1;
my $distrib_file = "";
my $tree_file = "";
my $wag = "";
my $d_scale = 1;
my $nCatG = 8;
my $avrg = 1;
my $alpha = 0.6;
for my $i (0 .. scalar(@ARGV)-1){
	if($ARGV[$i] eq "-i"){ $gene_name = $ARGV[$i+1];}
	elsif ($ARGV[$i] eq "-t") { 
		if($ARGV[$i+1] > 0 && $ARGV[$i+1] <= 1){
			$posterior_threshold = $ARGV[$i+1]; 
		}
		else{
			print STDERR "Posterior Threshold must be between 0 and 1\n";
			exit;
		}
	}
	elsif ($ARGV[$i] eq "-q"){ $QUIET=1; }
	elsif ($ARGV[$i] eq "-start"){ $START=$ARGV[$i+1]; }
	elsif ($ARGV[$i] eq "-stop"){ $STOP=$ARGV[$i+1]; }
	if($ARGV[$i] eq "-l"){ $r_length = $ARGV[$i+1];}
	if($ARGV[$i] eq "-max"){$max_output = $ARGV[$i+1];}
	if($ARGV[$i] eq "-min"){$min_output = $ARGV[$i+1];}
	if($ARGV[$i] eq "-w"){$local_window = $ARGV[$i+1];}
	if($ARGV[$i] eq "-cpu"){$cpu = $ARGV[$i+1];}
	if($ARGV[$i] eq "-f"){$flank =  $ARGV[$i+1];}
	if($ARGV[$i] eq "-debug"){$debug =  1;}
	if($ARGV[$i] eq "-distrib"){$distrib_file =  $ARGV[$i+1];}
	if($ARGV[$i] eq "-tree"){$tree_file =  $ARGV[$i+1];}
	if($ARGV[$i] eq "-weight"){$conserve_weight =  $ARGV[$i+1];}
	if($ARGV[$i] eq "-distance") {$d_scale = $ARGV[$i+1];}
	if($ARGV[$i] eq "-wag"){$wag = "-wag";}
	if($ARGV[$i] eq "-k"){$nCatG = $ARGV[$i+1];}
	if($ARGV[$i] eq "-alpha"){$alpha = $ARGV[$i+1];}
	if($ARGV[$i] eq "-avrg"){$avrg = $ARGV[$i+1];}
}


my @facts;

if(-e "$gene_name" && -e "$distrib_file"){
	
	my $protein_count = -1;
	open(ALIGNMENT,"$gene_name");
	while(my $line = <ALIGNMENT>){
		chomp($line);
		if($line =~ m/\>/){
			push(@gene_IDs,$line);
			++$protein_count;
			$protein[$protein_count]{'gene_ID'} = $line;
		}
		else{
			$protein[$protein_count]{'sequence'} .= $line;
		}
		$line =~ s/\s/+/g;
		if($offset_flag == 1){
			for(my $i = 0;$i < length($line);++$i){
				if(substr($line,$i,1) eq "-"){
					++$offset;	
				}
				$offset_array[$i] = $offset;
			}
			$offset_flag = 0;
		}
		if($line =~ m/Scer/){
			$offset_flag = 1;	
		}
	}
	++$protein_count;
	for(my $i = 0;$i < $protein_count;++$i){
		if($protein[$i]{'sequence'} ne ""){
			$p_length = length($protein[$i]{'sequence'});
			last;
		}
	}
	if($STOP == 0){
		$STOP = $p_length;	
	}
	
	close(ALIGNMENT);
	
	# Take the conserved state local distribution and put in array.
	open(LOCAL,$distrib_file) or die;
	my @all_sub_seq2;
	while(my $lines = <LOCAL>){
		chomp($lines);
		if($lines =~ m/^([0-9]+)\t([0-9\.]+)/){
			push(@all_sub_seq2,$2);
		}
	}
	
	local $temp = generate_random_string(10);
	local $temp2 = generate_random_string(10);
	
	open(TEMP2,">$temp2") or die;
	open(TEMP,">$temp") or die; 
	for(my $i = 0;$i < scalar(@all_sub_seq2);++$i){
		print TEMP ($i+1);
		print TEMP "\t";
		print TEMP2 ($i+1);
		print TEMP2 "\t";
		print TEMP $avrg;
		if($all_sub_seq2[$i] < $avrg/$conserve_weight){
			print TEMP2 $all_sub_seq2[$i];
		}
		else{
			print TEMP2 ($avrg/$conserve_weight);
		}
		
		print TEMP "\n";
		print TEMP2 "\n";
	}
	close(TEMP);
	close(TEMP2);
	
	print "Posterior threshold = $posterior_threshold\n";
	print "No local rate correction.\n";
	print "Expected length = $r_length\n";
	print "Conservation weight = $conserve_weight\t";
	print "Gap weight = $conserve_weight\n";
		
	$distance_fast = 1;
	
	# We will thread the maximal_likelihood code here because the C threading implementation is slow.
	my $spread_process = int(($STOP-($START-1)) / $cpu);
	for(my ($p) = $START;$p < ($STOP + 1);$p=$p+$spread_process+1){
		if($cpu == 1){
			$threads[$p] = run_ML($p,min($p+$spread_process,$STOP),$temp,20);
		}
		else{
			$threads[$p] = threads->create(\&run_ML,$p,min($p+$spread_process,$STOP),$temp,20);		
		}
	}
	for(my ($p) = $START;$p < ($STOP + 1);$p=$p+$spread_process+1){
		my $print_list;
		if($cpu == 1){
			$print_list = $threads[$p];	
		}
		else{
			$print_list = $threads[$p]->join();
		}
		my @pl = split(/\n/, $print_list);
		for(my $i = 0;$i < scalar(@pl);++$i){
			my $line = $pl[$i];
			chomp($line);
			if($line =~ m/([0-9]+)\t.+\t?.+\t([0-9\.-]+)/){
				$e{0}{$1} = $2;	
			}
		}
	}
		
	for(my ($p) = $START;$p < ($STOP + 1);$p=$p+$spread_process+1){
		if($cpu == 1){
			$threads[$p] = run_ML($p,min($p+$spread_process,$STOP),$temp2,$alpha);
		}
		else{
			$threads[$p] = threads->create(\&run_ML,$p,min($p+$spread_process,$STOP),$temp2,$alpha);	
		}
	}
	for(my ($p) = $START;$p < ($STOP + 1);$p=$p+$spread_process+1){
		my $print_list;
		if($cpu == 1){
			$print_list = $threads[$p];	
		}
		else{
			$print_list = $threads[$p]->join();
		}
		my @pl = split(/\n/, $print_list);
		for(my $i = 0;$i < scalar(@pl);++$i){
			my $line = $pl[$i];
			chomp($line);
			if($line =~ m/([0-9]+)\t.+\t?.+\t([0-9\.-]+)/){
				$e{1}{$1} = $2;	
			}
		}
	}
		
	unlink($temp);
	unlink($temp2);
	
	local $temp3 = generate_random_string(10);
	open(TEMP3,">$temp3") or die;
	for(my $i = 1;$i <= $p_length;++$i){
		$e{0}{$i} = $e{0}{$i};
		$e{1}{$i} = $e{1}{$i};
		
		
		print TEMP3 $e{0}{$i};
		print TEMP3 "\t";
		print TEMP3 $e{1}{$i};
		print TEMP3 "\n";
	}
	close(TEMP3);
	# Baum-Welch procedure
	open(BAUM,"$BW_exe -i $temp3 -e $r_length -l $p_length |") or die;
	my $expected_count = 0.01 * $p_length;
	my $expected_length = $r_length;
	while(my $lines = <BAUM>){
		if($lines =~ m/count\s:\s([0-9\.]+)/){
			$expected_count = $1;
		}
		if($lines =~ m/length\s:\s([0-9\.]+)/){
			$expected_length = $1;
		}
	}
	close(BAUM);
	unlink($temp3);
	# Initialize HMM parameters;
	print "Baum Welch count estimates : $expected_count (".sprintf("%.3f",$expected_count/$p_length)." per AA)\tlength estimates : $expected_length";
	if($expected_count/$p_length > 0.005){
		print "\t";
		print "Expected count too high, setting to ".(0.005  * $p_length)." (0.005 per AA).";
		$expected_count = 0.005 * $p_length;
	}
	if($expected_length < 30 || $expected_length < $min_output){
		if(30 < $min_output){
			print "\t";
			print "Expected length too low, setting to $min_output.";
			$expected_length = $min_output;
		}
		else{
			print "\t";
			print "Expected length too low, setting to 30.";
			$expected_length = 30;
		}
	}
	
	print "\n";
	print "\n";
	if($expected_count < 1){
		$expected_count = 1;	
	}
	
	# Transition Frequencies;
	$a[0][1] = $expected_count/$p_length;
	$a[0][0] = 1-$a[0][1];
	
	$a[1][0] = 1/$expected_length;
	$a[1][1] = 1-$a[1][0];
	
	# Initiation probabilities;
	$begin[0] = $a[0][0];
	$begin[1] = 1-$begin[0];
	
	# Process of forward algorithm, 1 index;
	for(my $i = 1;$i<=$p_length;++$i){
		
		#Go through the states;
		for(my $j = 0;defined $a[$j];++$j){
			my $prev_f;
			
			#Set initial values for Forward
			if($i == 1){
				$f[$j][$i] = $begin[$j];	
			}
			else{
				for(my ($k) = 0;(defined $a[$k]);++$k){
					$prev_f += $f[$k][$i-1] * $a[$k][$j];
				}
				$f[$j][$i] = $prev_f;
			}
			
			$f[$j][$i] *= exp($e{$j}{$i});
			$f_scaling[$i] += $f[$j][$i];
		}
		for(my ($j) = 0;defined $a[$j];++$j){
			$f[$j][$i] = $f[$j][$i] / $f_scaling[$i];	
		}
		$log_fs_process += log($f_scaling[$i]);
		$log_fs[$i] = $log_fs_process;	
	}	
	
	# Forward Termination
	my $P_x;
	for(my ($k) = 0;defined $a[$k];++$k){
		$P_x += $f[$k][$p_length];
	}
	$P_x = log($P_x) + $log_fs[$p_length];
	
	# Initiation of Backward Algorithm [ Scaled ]	
	for(my ($k) = 0;defined $a[$k];++$k){
		$b[$k][$p_length] = 1/$f_scaling[$p_length];
	}
	
	# Implementation of Backward Algorithm
	for($i = ($p_length-1);$i >= 1;$i = $i-1){
		
		#Scaling values of b
		$log_bs_process += log($f_scaling[$i+1]);
		$log_bs[$i+1] = $log_bs_process;
		
		for(my ($k) = 0;defined $a[$k];++$k){
			for(my ($j) = 0;defined $a[$j];++$j){
				$b[$k][$i] += ($b[$j][$i+1] * exp($e{$j}{$i+1}) * $a[$k][$j])/$f_scaling[$i];
			}
		}
	}
	
	#Termination of Scaling Values
	$log_bs_process += log($f_scaling[1]);
	$log_bs[1] = $log_bs_process;
	
	#Termination step of Forward Algorithm [ Scaled Value ]
	my ($f_P);
	for($k = 0;defined $f[$k];++$k){
		$f_P += $f[$k][$p_length];	
	}
	
	#Termination step of Backward Algorithm [ Scaled Value ]	
	my ($b_P);
	for($k = 0;defined $b[$k];++$k){
		$b_P += exp($e{$k}{1}) * $b[$k][1] * $begin[$k];
	}
	
	#Posterior in Log Space		
	my ($posterior_string) = "";
	my @posterior = ();
	
	for($i = 1;$i <= $p_length;$i = $i + 1){
		#$posterior_state = which posterior state to threshold.
		my $posterior_state;
		$posterior_state = 0;
		
		if($f[$posterior_state][$i] == 0 || $b[$posterior_state][$i] == 0){
			$posterior[$i] = -10**16;
		}
		else{
			$posterior[$i] = log($f[$posterior_state][$i]) + $log_fs[$i] + log($b[$posterior_state][$i]) + $log_bs[$i] - $P_x;
			
		}
		
		if(1-exp($posterior[$i]) > $posterior_threshold){
			$posterior_string .= "1";	
		}
		else{
			$posterior_string .= "0";
		}
		
		$print_debug = $debug;
		if($print_debug == 1){
			print $i;
			print "\t";
			print (1-exp($posterior[$i]));
			print "\t";
			print substr($protein[0]{'sequence'},$i-1,1);
			print "\t";
			print $e{0}{$i};
			print "\t";
			print $e{1}{$i};
			print "\t";
			print $local_rate[$i-1];
			print "\t";
			
			print $all_sub_seq2[$i-1];	
			
			#print ($local_rate[$i]/$average_sub);
			print "\n";
		}
	}
	my %pstarts; my %pstops; my $pmatchn=0;
	while ($posterior_string=~m/([1]+)/g) { 
		my $test=$1;
		# Limit motif search to 20 in size.
		if(length($test) > $max_output && $max_output != 0){ next;}
		if(length($test) < $min_output){next;}
		my $pos=pos($posterior_string);
		$pmatchn++;
		$pstarts{$pmatchn}= $pos-length($test);
		$pstops{$pmatchn}=$pstarts{$pmatchn}+length($test);
	}
	my $site_print = 1;
	for my $m (1 .. $pmatchn) {
	
		my $temp_array_pos;
		my $min_temp_array = 1;
		
		my $temp_left_array_pos;
		my $max_left_array = 1;
		
		my $temp_right_array_pos;
		my $max_right_array = 1;
		
		my $average_posterior = 0;
		
		for my $l ($pstarts{$m} .. $pstops{$m}-1){
			#push(@temp_array,exp($posterior[$l]));	
			if(exp($posterior[$l+1]) < $min_temp_array){
				$min_temp_array = exp($posterior[$l+1]);
				$temp_array_pos = $l;
			}
			$average_posterior += (1-exp($posterior[$l+1]));
		}

		for my $l (max(1,$pstarts{$m} - $local_window) .. $pstarts{$m}-1){
			if(exp($posterior[$l]) < $max_left_array){
				$max_left_array = exp($posterior[$l]);
				$temp_left_array_pos = $l;
			}
		}
		for my $l ($pstops{$m}+1 .. min($p_length,$pstops{$m}+$local_window)){
			if(exp($posterior[$l]) < $max_right_array){
				$max_right_array = exp($posterior[$l]);
				$temp_right_array_pos = $l;
			}	
		}
		$max_in_string = 1-$min_temp_array;
		
		if($max_in_string < $posterior_threshold){next;}
		$left_string = 1-$max_left_array;
		$right_string = 1-$max_right_array;
		
		print "Site #$site_print : ".($pstarts{$m}+1)."-".$pstops{$m}."\t$max_in_string";
		print "\n";
		for(my $i = 0;$i<$protein_count;++$i){
			my $string_to_print = substr($protein[$i]{'sequence'},$pstarts{$m},$pstops{$m}-$pstarts{$m});
			if($string_to_print eq ""){next;}
			print $protein[$i]{'gene_ID'};
			print "\t";
			print "\t";
			
			if($flank > 0){
				print substr($protein[$i]{'sequence'},$pstarts{$m}-$flank,$flank);
				print "  ";
			}
			
			print substr($protein[$i]{'sequence'},$pstarts{$m},$pstops{$m}-$pstarts{$m});	
			
			if($flank > 0){
				print "  ";
				print substr($protein[$i]{'sequence'},$pstops{$m},$flank);		
			}
			
			print "\n";
		}
		print "\n";
		++$site_print;
	}	
	if($pmatchn > 0){
		print "Finished analyzing $gene_name. Found ".($site_print-1)." sites.\n";	
		print "======================================================\n";
	}
}
else{
	print "Couldn't find the alignment file.\n";	
}
sub generate_random_string
{
	my $length_of_randomstring=shift;# the length of 
			 # the random string to generate

	my @chars=('a'..'z','A'..'Z','0'..'9','_');
	my $random_string;
	foreach (1..$length_of_randomstring) 
	{
		# rand @chars will generate a random 
		# number between 0 and scalar @chars
		$random_string.=$chars[rand @chars];
	}
	return $random_string;
}

sub max {
	my($max) = shift(@_);

	foreach $temp (@_) {
		$max = $temp if $temp > $max;
	}
	return($max);
}
sub min {
	my($min) = shift(@_);

	foreach $temp (@_) {
		$min = $temp if $temp < $min;
	}
	return($min);
}
sub average {
	my $average;
	foreach my $temp (@_) {
		$average += $temp;
	}
	$average = $average / scalar(@_);
	return($average);
}
sub compute_local{
	my $window = @_[0];

	my @return_array;
	# Compute the local rate of substitution and put into array.
	for(my $j = 0;$j < scalar(@sub_array);++$j){
		my $min_rate = 100;

		for(my $i = -$window/2 + $j;$i <= $window/2 + $j;++$i){
			if($i < 0){ next;}
			if($sub_array[$i] != 0){
				if($sub_array[$i] < $min_rate){
					$min_rate = $sub_array[$i];
				}
			}
						
		}
		push(@return_array,$min_rate);
	}
	return @return_array;
}
sub run_ML{
	my $s1 = @_[0];
	my $s2 = @_[1];
	my $distrib = @_[2];
	my $alpha = $_[3];
	my $distrib_cmd;
	if($distrib eq ""){
		$distrib_cmd = "";
	}
	else{
		$distrib_cmd = "-distrib $distrib ";	
	}
	
	my @print_array;
	my $command = "$ML_exe -i $gene_name -tree $tree_file -distance $d_scale $wag -start $s1 -stop $s2 $distrib_cmd -alpha $alpha -k $nCatG |";
	open(CML,$command) or die "$!";
	while(my $line = <CML>){
		push(@print_array,$line);
	}
	close(CML);
	return join("",@print_array);
}

sub interrupt {
    print STDERR "\nExiting script, cleaning up created files.\n";
    if(-e "$temp"){
    	    unlink($temp);
    }
     if(-e "$temp2"){
    	    unlink($temp2);
    }
    if(-e "$temp3"){
    	    unlink($temp3);
    }
    exit;
}
