// ##############################################################################
// Maximal Likelihood algorithm. Not for distribution.
// Copyright Alex Nguyen Ba 2011
// Version 1.2
//
// LICENCE 
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ##############################################################################


using namespace std;

#include <string>
#include <iostream>
#include <fstream>
#include <stdlib.h> 
#include <cstring>
#include <math.h>
#include <sstream>
#include <vector>
#include <acml.h>
#include <boost/regex.hpp>

double P[20][20] = { 0 };
double P1[20][20] = { 0 };

double P_gap[2][2] = {0};
double P1_gap[2][2] = {0};

double *lambda = new double[26];
double *lambda_gap = new double[2];

//###############################################################################
// Symbols (Amino acids) Change to only ATCG if DNA will dramatically improve
// speed.
//###############################################################################

string aalist[20] = {"L","S","K","I","E","N","T","D","V","A","G","R","F","P","Q","Y","H","M","C","W"};
int aalist_size = sizeof(aalist)/sizeof(string);

string aalist_gap[2] = {"X","-"};
int aalist_size_gap = sizeof(aalist_gap)/sizeof(string);

class Sequence {
	
private:
	string m_seq_ID;
	string m_sequence;
public:
	
	//Constructor

	//Sequence(){
	//	m_seq_ID = NULL;
	//	m_sequence = NULL;
	//}
	
	//Setters
	bool appendLine(string seq_line){
		m_sequence += seq_line;
	}
	
	bool setHeader(string header){
		m_seq_ID = header;	
	}
	
	//Getters
	string getHeader(){
		return m_seq_ID;	
	}
	
	string getSequence(){
		return m_sequence;	
	}
};

class Genome {
private:
	int m_genome_size; //Number of sequences	
	map<int , Sequence*> list;
public:
	// Constructor
	Genome(){
		m_genome_size = -1;	
	}
	
	// Getter
	int getGenomeSize(){
		return m_genome_size+1;	
	}
	
	// Main function used.
	bool read_fasta(string fastafile){
		ifstream file;
		file.open(fastafile.c_str());
		int i =0;
		if (!file) return 0;
		string gene_name;
		string line;
		
		while( getline( file, line ) ){
			if (line[0] == '>') {
				++m_genome_size;
				string header;
				header = "";
				for(int j = 1;j < line.length();j++){
					header += line[j];	
				}
				list[m_genome_size] = new Sequence;
				list[m_genome_size]->setHeader(header);
			}
			else{
				if(line.rfind("*")!=string::npos){
					line = line.erase(line.rfind("*"),1);	
				}
				if(line.rfind("\n")!=string::npos){
					line = line.erase(line.rfind("\n"),1);	
				}
				if(line.rfind("\r")!=string::npos){
					line = line.erase(line.rfind("\r"),1);	
				}
				
				list[m_genome_size]->appendLine(line);
			}
		}	
		
		return 1;	
	}
	
	// Grab a child sequence.
	Sequence getSequenceByID(int ID){
		return *list[ID];	
	}
	//Delete a child sequence
	void deleteSequenceByID(int ID){
		for(int i = ID;i < m_genome_size;++i){
			//delete[](list[i]);
			list[i] = list[i+1];
		}
		//delete[](list[m_genome_size]);
		list.erase(m_genome_size);
		m_genome_size -= 1;
	}
};
class Node{
private:
	string m_sequence;
	string m_node_ID;
	bool is_parent;
	int child_1;
	int child_2;
	double distance_1;
	double distance_2;
	string species;
public:
	Node(){
		is_parent = 0;	
	}
	// Getters
	bool isParent(){
		return is_parent;	
	}
	int getChild_1(){
		return child_1;	
	}
	int getChild_2(){
		return child_2;	
	}
	double getDistance_1(){
		return distance_1;	
	}
	double getDistance_2(){
		return distance_2;	
	}
	string getSpecie(){
		return species;	
	}
	// Setters
	bool appendLine(string seq_line){
		m_sequence += seq_line;
	}
	bool set_parent(){
		is_parent = 1;
	}
	bool set_child_1(int childID){
		child_1 = childID;	
	}
	bool set_child_2(int childID){
		child_2 = childID;	
	}
	bool set_distance_1(double d){
		distance_1 = d;	
	}
	bool set_distance_2(double d){
		distance_2 = d;	
	}
	string getSequence(){
		return m_sequence;	
	}
	bool set_specie(string specie){
		species = specie;	
	}	
};
class NodeList{
private:
	int Nodelist_size;
	vector<Node*> list;
	
public:
	// Constructor
	NodeList(){
		Nodelist_size = -1;
	}
	
	bool createNode(){
		++Nodelist_size;
		list[Nodelist_size] = new Node;
	}
	bool resize_vector(int a){
		list.resize(2 * a); 
	}
	// Grab a node.
	Node getNodeByID(int ID){
		return *list[ID];	
	}
	
	// Getter
	int getNodeListSize(){
		return Nodelist_size+1;	
	}
	//
	void setSequenceOfNode(int ID,string seq){
		list[ID]->appendLine(seq);
	}
	
	//
	void setChildAsParent(int ID){
		list[ID]->set_parent();	
	}
	
	void setChild1(int ID,int childID){
		list[ID]->set_child_1(childID);	
	}
	void setChild2(int ID,int childID){
		list[ID]->set_child_2(childID);	
	}
	void setDistance1(int ID,double d){
		list[ID]->set_distance_1(d);	
	}
	void setDistance2(int ID,double d){
		list[ID]->set_distance_2(d);	
	}
	void setSpeciebyID(int ID,string specie){
		list[ID]->set_specie(specie);	
	}
	string print_tree(int ID){
		
		
		string string_left;
		string string_right;
		if(list[ID]->isParent() == 0){
			return list[ID]->getSpecie();
			
		}
		else{
			int child_1 = list[ID]->getChild_1();
			int child_2 = list[ID]->getChild_2();
			
			stringstream out;
			out.precision(5);
			out << fixed << list[ID]->getDistance_1();
			stringstream out2;
			out2.precision(5);
			out2 << fixed << list[ID]->getDistance_2();
			
			
			string_left = (this->print_tree(child_1) + ":" + out.str());
			string_right = (this->print_tree(child_2)  + ":" + out2.str());
			out.clear();
			out2.clear();
		}
		
		
		
		return("(" + string_left + "," + string_right + ")");
	}
};
int count_in_string(string haystack, string needle){
	int count_in_string = 0;
	int i = 0;
	while(haystack.find(needle,i)!=string::npos){
		i = haystack.find(needle,i)+1;
		++count_in_string;
	}
	return count_in_string;
}

string help(){
	string helpstr;
	helpstr = "You are using Maximal Parsimony v.1.1 copyright Alex Nguyen Ba 2009\n";
	helpstr += "-i input file\n";
	helpstr += "-start [optional] Specify the position to start the algorithm in the gene.\n";
	helpstr += "-stop [optional] Specify the position to stop the algorithm in the gene.\n";
	helpstr += "-gap Specify if gaps are ignored.\n";
	return(helpstr);	
}

string convert(int number){
	ostringstream sin;
	sin << number;
	string val = sin.str();
	return(val);
}
double gammln(double xx){
	double x,y,tmp,ser;
	double coff[6];
	coff[0] = 76.18009172947146;
	coff[1] = -86.50532032941677;
	coff[2] = 24.01409824083091;
	coff[3] = -1.231739572450155;
	coff[4] = 0.1208650973866179e-2;
	coff[5] = -0.5395239384953e-5;

	y = x = xx;

	tmp = x + 5.5;
	tmp -= (x + 0.5)*log(tmp);

	ser = 1.000000000190015;
	for(int j = 0;j <= 5;++j){
		ser += coff[j]/++y;
	}
	return (-1 * tmp + log(2.5066282746310005*ser/x));

}
double alnorm(double x, int upper){
	double a1 = 5.75885480458; 
  	double a2 = 2.62433121679; 
  	double a3 = 5.92885724438; 
  	double b1 = -29.8213557807; 
  	double b2 = 48.6959930692; 
  	double c1 = -0.000000038052; 
  	double c2 = 0.000398064794; 
  	double c3 = -0.151679116635; 
  	double c4 = 4.8385912808; 
  	double c5 = 0.742380924027; 
  	double c6 = 3.99019417011;
 	double con = 1.28;
  	double d1 = 1.00000615302;
  	double d2 = 1.98615381364;
  	double d3 = 5.29330324926;
  	double d4 = -15.1508972451;
  	double d5 = 30.789933034;
  	double ltone = 7.0;
  	double p = 0.39894228044; 
  	double q = 0.39990348504;
  	double r = 0.398942280385;
  	double utzero = 18.66;

  	double up = upper;
  	double z = x;

	if(z < 0){
		if (up){
			up = 0;
		}
		else{
			up = 1;
		}
		z *= -1;
	}
	if(ltone < z && (!(up) || utzero < z)){
		if(up){
			return 0.0;
		}
		else{
			return 1.0;
		}
	}

	double y = 0.5 * z * z;
	double value;
	if(z <= con){
		value = 0.5 - z * (p - q * y / (y + a1 + b1 / (y + a2 + b2 / (y + a3 ))));
	}
	else{
		value = r * exp (-1 * y) / (z + c1 + d1 / (z + c2 + d2 / (z + c3 + d3 / (z + c4 + d4 / (z + c5 + d5 / (z + c6 ))))));
	}
	if(!(up)){
		value = 1.0 - value;
	}
	return value;
}

double ppnd7(double p){
	double normal_dev;

	double q;
	double r;

	double a0 = 3.3871327179;	double a1 = 5.0434271938E+01;
	double a2 = 1.5929113202E+02;	double a3 = 5.9109374720E+01;
	double b1 = 1.7895169469E+01;	double b2 = 7.8757757664E+01;
	double b3 = 6.7187563600E+01;

	double c0 = 1.4234372777;	double c1 = 2.7568153900;
	double c2 = 1.3067284816;	double c3 = 1.7023821103E-01;
	double d1 = 7.3700164250E-01;	double d2 = 1.2021132975E-01;

	double e0 = 6.6579051150;	double e1 = 3.0812263860;
	double e2 = 4.2868294337E-01;	double e3 = 1.7337203997E-02;
	double f1 = 2.4197894225E-01;	double f2 = 1.2258202635E-02;

	q = p - 0.5;
	if(fabs(q) <= 0.425){
		r = 0.180625 - q * q;
		normal_dev = q * (((a3 * r + a2) * r + a1) * r + a0) / ((( b3 * r + b2) * r + b1) * r + 1);	
	}
	else{
		if (q < 0){
			r = p;
		}
		else{
			r = 1 - p;
		}
		if(r <= 0){
			return 0;
		}
		r = sqrt(-log(r));
		if(r <= 5){
			r = r - 1.6;
			normal_dev = (((c3 * r + c2) * r + c1) * r + c0) / ((d2 * r + d1) * r + 1);
		}
		else{
			r = r - 5;
			normal_dev = (((e3 * r + e2) * r + e1) * r + e0) / ((f2 * r + f1) * r + 1);
		}
		if(q < 0){
			normal_dev *= -1;
		}

	}
	return normal_dev;
}
double gammain(double x, double p){
	// Returns the incomplete gamma ratio I(x,alpha) where x is the upper limit of the integration and alpha is the shape parameter. Appears to be more stable than gammad
	// AS32
	double g = gammln(p);
	
	double ACU = 1e-12;
	double OFLO = 1e30;
	double gin = 0,rn = 0, a = 0, b = 0, an = 0, dif = 0, term = 0;
	double pn[6] = {0};

	if(x == 0){
		return 0;
	}
	if(x < 0 || p <= 0){
		return -1;
	}
	double factor = exp(p*log(x) - x - g);
	if(x > 1 && x >= p){
		a = 1 - p;
		b = a + x + 1;
		term = 0;
		pn[0] = 1;
		pn[1] = x;
		pn[2] = x+ 1;
		pn[3] = x * b;
		gin = pn[2]/pn[3];
	}
	else{
		gin = 1;
		term = 1;
		rn = p;
		while(term > ACU){
			rn += 1;
			term *= x / rn;
			gin += term;
		}	
		gin *= factor / p;
		return gin;
	}
	double it = 0;
	int i = 0;
	while(it < 100){
		++it;
		a += 1;
		b += 2;
		term += 1;
		an = a * term;
		for(i = 0;i < 2;++i){
			pn[i+4] = b * pn[i+2] - an * pn[i];
		}
		if(pn[5] != 0){
			rn = pn[4]/pn[5];
			dif = abs(gin - rn);
			if(dif > ACU){
				gin = rn;
			}
			else if(dif <= ACU * rn){
				break;
			}
		}
		for(i = 0;i < 4;++i){
			pn[i] = pn[i+2];
		}
		if(abs(pn[4]) < OFLO){
			continue;
		}
		for(i = 0;i < 4;++i){
			pn[i] /= OFLO;
		}
	}
	gin = 1 - factor * gin;
	return gin;
}
double gammad(double x, double p){
	double pn1,pn2,pn3,pn4,pn5,pn6,arg,c,rn,a,b,an;
	double OFLO = 1e37;
	double fn_val = 0;
	double tol = 1e-14;
	if(p <= 0 || x < 0){
		cout << "ERROR" << p << "	" << x << endl;
		return 0;
	}
	if(x == 0){
		return 0;
	}
	if(p > 1000){
		pn1 = 3 * sqrt(p) * (  pow((x / p),1.0/3.0) + 1/(9*p) - 1   );
		fn_val = alnorm(pn1,0);
		return fn_val;
	}

	if(x > 1e08){
		return 1;
	}

	if(x <= 1 || x < p){
		arg = p * log(x) - x - gammln(p + 1);
		c = 1;
		fn_val = 1;
		a = p;
		while(1){
			a += 1;
			c *= x / a;
			fn_val += c;
			if(c <= 1e-14){
				break;
			}
		}
		arg += log(fn_val);
		fn_val = 0;
		if(arg >= -88){
			fn_val = exp(arg);
		}
	}
	else{
		arg = p * log(x) - x - gammln(p);
		a = 1 - p;
		b = a + x + 1;
		c = 0;
		pn1 = 1;
		pn2 = x;
		pn3 = x + 1;
		pn4 = x * b;
		fn_val = pn3/pn4;
		int max_it = 100;
		int it = 0;
		while(it < max_it){
			++it;
			a += 1;
			b += 2;
			c += 1;
			an = a * c;
			pn5 = b * pn3 - an * pn1;
			pn6 = b * pn4 - an * pn2;
			if(fabs(pn6) > 0){
				rn = pn5 / pn6;
				if(rn < 1){
					if(fabs(fn_val - rn) <= tol * rn){
						break;
					}
				}
				else{
					if(fabs(fn_val - rn) <= tol){
						break;
					}
				}
				fn_val = rn;
			}

			pn1 = pn3;
			pn2 = pn4;
			pn3 = pn5;
			pn4 = pn6;
			if(fabs(pn5) >= OFLO){
				pn1 /= OFLO;
				pn2 /= OFLO;
				pn3 /= OFLO;
				pn4 /= OFLO;
			}
		}

		arg = arg + log(fn_val);
		fn_val = 1;
		if(arg >= -88){
			fn_val = 1 - exp(arg);
		}
	}

	return fn_val;
}
double ppchi2(double p, double v){
	
	double aa = 0.6931471806;
	double c1 = 0.01;	double c2 = 0.222222;	double c3 = 0.32;	double c4 = 0.4;
	double c5 = 1.24;	double c6 = 2.2;	double c7 = 4.67;	double c8 = 6.66;
	double c9 = 6.73;	double c10 = 13.32;	double c11 = 60.0;	double c12 = 70.0;
	double c13 = 84.0;	double c14 = 105.0;	double c15 = 120.0;	double c16 = 127.0;
	double c17 = 140.0;	double c18 = 175.0;	double c19 = 210.0;	double c20 = 252.0;
	double c21 = 264.0;	double c22 = 294.0;	double c23 = 346.0;	double c24 = 420.0;
	double c25 = 462.0;	double c26 = 606.0;	double c27 = 672.0;	double c28 = 707.0;
	double c29 = 735.0;	double c30 = 889.0;	double c31 = 932.0;	double c32 = 966.0;
	double c33 = 1141.0;	double c34 = 1182.0;	double c35 = 1278.0;	double c36 = 1740.0;
	double c37 = 2520.0;	double c38 = 5040.0;
	double e = 0.5E-06;
	double maxit = 20;
	double pmax = 0.999998;
	double pmin = 0.000002;
	double x;
	double value = -1;
	if(p<pmin || pmax<p){
		cout << "ERROR p<min or p<max" << endl;
		return 0;
	}
	if(v<=0){
		cout << "ERROR v<=0" << endl;
		return 0;
	}
	
	double xx = 0.5 * v;
	double c = xx - 1;
	double q,p1,p2,t,ch,a,temp,s1,s2,s3,s4,s5,s6;
	double g = gammln(v/2);
	
	if(v < -1 * c5 * log(p)){
		ch = pow((p*xx*exp(g+xx*aa)),1.0/xx);
		if(ch < e){
			return ch;
		}
	}
	else if(v <= c3){
		ch = c4;
		a = log(1.0-p);
		
		while(1){
			q = ch;
			p = 1 + ch * (c7 + ch);
			p2 = ch * (c9 + ch * (c8 + ch));
			t = -0.5 + (c7 + 2 * ch) / p1 - (c9 + ch * (c10 + 3 * ch))/p2;
			ch -= (1 - exp(a + g + 0.5 * ch + c * aa) * p2/p1) / t;
			
			if(fabs(q/ch - 1) <= c1){
				break;
			}
		}
	}
	else{
		x = ppnd7(p);
		p1 = c2/v;
		ch = v * pow((x * sqrt(p1) + 1 - p1),3.0);
		
		if(c6 * v + 6 < ch){
			ch = -2 * (log(1-p) - c * log(0.5*ch) + g);
		}
	}

	for(int i = 1;i < maxit;++i){
		q = ch;
		p1 = 0.5 * ch;	
		temp = gammain(p1,xx);
		p2 = p - temp;
		
		t = p2 * exp(xx * aa + g + p1 - c * log(ch));
		double b = t / ch;
		a = 0.5 * t - b * c;


		s1 = (c19 + a * (c17 + a * (c14 + a * (c13 + a * (c12 + c11 * a))))) / c24;
		s2 = (c24 + a * (c29 + a * (c32 + a * (c33 + c35 * a)))) / c37;
		s3 = (c19 + a * (c25 + a * (c28 + c31 * a))) / c37;
		s4 = (c20 + a * (c27 + c34 * a) + c * (c22 + a * (c30 + c36 * a))) / c38;
		s5 = (c13 + c21 * a + c * (c18 + c26 * a)) / c37;
		s6 = (c15 + c * (c23 + c16 * c)) / c38;
		ch = ch + t * (1.0 + 0.5 * t * s1 - b * c * (s1 - b * (s2 - b * (s3 - b * (s4 - b * (s5 - b * s6))))));
		
		if(e < fabs(q/ch - 1)){
			return ch;
		}
	}
	
	return ch;
}
double ppgamma(double p,double alpha){
	return (ppchi2(p,2.0*alpha)/(2.0*alpha));
}
double ratei(int i,int k,double alpha){
	double a;
	double b;
	if(i == 1){
		a = 0.0;
	}
	else{
		a = ppgamma(((double)i-1.0)/k,alpha);
	}
	if(i == k){
		return ((double)k * (1 - gammain(a * alpha,alpha+1)));
	}
	else{
		b = ppgamma(((double)i/k),alpha);
		return ((double)(gammain(b * alpha, alpha+1) - gammain(a * alpha, alpha + 1))*(double)k);
	}
}

int main(int argc, char * argv[]) {
	
	// ##################################################
	// Usage
	// ##################################################
	
	cout.precision(20);
	string fastafile;
	int start = 1;
	int stop = 0;
	double d_scale = 1;
	string distrib_file = "";
	string treefile;
	int wag = 0;
	int nCatG = 8;
	double alpha = 1.0;
	if(argc<2){
		cout << "usage: -i fastafile ... or try -help for more information." << endl;
		return 0;
	}
	string inputs ("-i");
	string starts ("-start");
	string stops ("-stop");
	string distances ("-distance");
	string distribs ("-distrib");
	string trees ("-tree");
	string wags ("-wag");
	string alphas ("-alpha");
	string nCatGs ("-k");
	
	for(int i = 1;i < argc;++i){
		if(argv[i] == inputs){
			fastafile = argv[i+1];
		}
		if(argv[i] == starts){
			start = atoi(argv[i+1]);
		}
		if(argv[i] == stops){
			stop = atoi(argv[i+1]);
		}
		if(argv[i] == distances){
			d_scale = atof(argv[i+1]);
		}
		if(argv[i] == distribs){
			distrib_file = argv[i+1];
		}
		if(argv[i] == trees){
			treefile = argv[i+1];	
		}
		if(argv[i] == wags){
			wag = 1;	
		}
		if(argv[i] == alphas){
			alpha = atof(argv[i+1]);
		}
		if(argv[i] == nCatGs){
			nCatG = atoi(argv[i+1]);
		}
	}
	
	if(fastafile.length() == 0){ 
		cout << "Requires a fasta file." << endl;
		return 0;
	}
	int letter_order[26];
	letter_order['L' - 'A'] = 0;
	letter_order['S' - 'A'] = 1;
	letter_order['K' - 'A'] = 2;
	letter_order['I' - 'A'] = 3;
	letter_order['E' - 'A'] = 4;
	letter_order['N' - 'A'] = 5;
	letter_order['T' - 'A'] = 6;
	letter_order['D' - 'A'] = 7;
	letter_order['V' - 'A'] = 8;
	letter_order['A' - 'A'] = 9;
	letter_order['G' - 'A'] = 10;
	letter_order['R' - 'A'] = 11;
	letter_order['F' - 'A'] = 12;
	letter_order['P' - 'A'] = 13;
	letter_order['Q' - 'A'] = 14;
	letter_order['Y' - 'A'] = 15;
	letter_order['H' - 'A'] = 16;
	letter_order['M' - 'A'] = 17;
	letter_order['C' - 'A'] = 18;
	letter_order['W' - 'A'] = 19;
	//###############################################################################
	// Handle the input fasta file.
	//###############################################################################
	Genome gen;
	if(!gen.read_fasta(fastafile)){
		cout << "Cannot read the fasta file." << endl;	
		return 0;
	}
	
	//###############################################################################
	// Sequence mapping
	// The code must recognize the following:
	// For a specie in the tree, what is the corresponding label in the fasta file.
	// By default, the mapping is the specie in the fasta file to the tree file.
	//###############################################################################
	int number_of_species = gen.getGenomeSize();
	
	string species_tree[number_of_species];
	string species_fasta[number_of_species];
	
	for(int i = 0;i < number_of_species;++i){
		string header = gen.getSequenceByID(i).getHeader();
		species_fasta[i] = header;
		species_tree[i] = header;
	}
	
	//###############################################################################
	// Phylogenetic tree in Newick format. 
	// Note, species not in this tree will not be parsed.
	// The tree will accomodate itself given the input file.
	// If the tree format is wrong, the code will likely fail.
	//###############################################################################
	
	ifstream file;
	file.open(treefile.c_str(),ifstream::in);
	if (!file){
		cout << "Can't find tree file." << endl;
		return 0;
	}
	string line = "";
	string tree = "";
	
	while(file >> line){
		tree += line;	
	}
	
	// Remove semi-colon.	
	if(tree.rfind(";")!=string::npos){
		tree = tree.erase(tree.rfind(";"),1);	
	}

	string print_tree = tree;
	//###############################################################################
	// Tree parsing
	// First step escapes all the tree_tags to regex format
	// Second step deletes empty sequences
	// Third step fills the node child array.
	// Fourth step removes unused species.
	// Last step fills parent array.
	//###############################################################################
	
	// Escape the tree string.
	for(int i = 0;i < number_of_species;++i){
		string::size_type pos = 0;
		if ( (pos = species_tree[i].find(".", pos)) != string::npos ) {
			species_tree[i].replace( pos, 1, "\\." );
			pos++;
		}
		pos = 0;
		if ( (pos = species_fasta[i].find(".", pos)) != string::npos ) {
			species_fasta[i].replace( pos, 1, "\\." );
			pos++;
		}
	}
	// Remove empty sequences.
	// Remove sequences not in the tree
	boost::regex pattern;
	for(int j = 0; j < gen.getGenomeSize();++j){
		pattern = gen.getSequenceByID(j).getHeader();
		if(!(regex_search(tree,pattern))){
			gen.deleteSequenceByID(j);
			--j;
			continue;
		}


		string my_sequence = gen.getSequenceByID(j).getSequence();
		int length_protein = my_sequence.length();
		
		if(length_protein == 0){
			gen.deleteSequenceByID(j);
			--j;
		}
	}
	boost::regex pattern2;
	std::string::const_iterator start_tree, end_tree;
	std::string::const_iterator start_line, end_line;
	boost::match_results<std::string::const_iterator> what; 
	boost::match_results<std::string::const_iterator> what2; 
	
	NodeList nodes;	
	nodes.resize_vector(number_of_species);
	// Populate the tree leaves into node IDs.
	for(int j = 0;j<gen.getGenomeSize();++j){
		string species_to_print;
		for(int i=0;i<number_of_species;++i){
			
			pattern = species_fasta[i];
			if(regex_search(gen.getSequenceByID(j).getHeader(),pattern)){
				pattern2 = species_tree[i];
				species_to_print = species_tree[i];
				stringstream out;
				out << j;
				tree = regex_replace(tree,pattern2,out.str());
				out.clear();
			}
		}
		nodes.createNode();
		
		nodes.setSpeciebyID(nodes.getNodeListSize()-1,species_to_print);
		
		nodes.setSequenceOfNode(nodes.getNodeListSize()-1,gen.getSequenceByID(j).getSequence());
	}

	
	start_tree = tree.begin(); 
	end_tree = tree.end();
	
	// Remove empty clades.
	pattern = "(.*)\\([A-Za-z\\.]+[0-9]*:[0-9\\.]+,[A-Za-z\\.]+[0-9]*:[0-9\\.]+\\):([0-9\\.]+)(.*)";
	while(regex_search(tree,what,pattern)){
		tree = what[1] + "REMOVE" + ":" + what[2] + what[3];
	}
	
	// Remove empty leaves.
	// Pattern 1 : ,REMOVE:0.1234)
	// Remove the pattern and a balanced parenthesis.
	start_tree = tree.begin(); 
	end_tree = tree.end();
	pattern = "(.*),[A-Za-z\\.]+[0-9]*:[0-9\\.]+\\)(.*)";
	while(regex_search(start_tree, end_tree, what, pattern)){
		// First part of the tree
		string first_part = what[1];
		// Last part of the tree
		string last_part = what[2];
		
		// We must remove a balanced parenthesis in the first part. 
		// Scan from right to left.
		int balanced = 0;
		for(int i = first_part.length()-1;i >= 0;--i){                                                     
			string letter = first_part.substr(i,1);
			
			if(letter == "("){
				if(balanced == 0){
					// Remove the encountered letter.
					tree = first_part.substr(0,i) + first_part.substr(i+1) + last_part;
					break;
				}
				--balanced;
			}
			else if(letter == ")"){
				++balanced;	
			}
		}
		pattern2 = "(.+:)([0-9\\.]+):([0-9\\.]+)(.+)";
		std::string::const_iterator start_tree2, end_tree2;
		
		start_tree2 = tree.begin();
		end_tree2 = tree.end();
		
		if(regex_search(start_tree2,end_tree2,what2,pattern2)){
			string first;
			string second;
			first = what2[2];
			second = what2[3];
			double sum = atof(first.c_str()) + atof(second.c_str());
			
			stringstream out;
			out << sum;
			tree = what2[1] + out.str() + what2[4];
			out.clear();
			
		}
		start_tree = tree.begin(); 
		end_tree = tree.end();
	}
	
	// Pattern 2 : (REMOVE:0.1234,
	// Remove the pattern and a balanced parenthesis.
	start_tree = tree.begin(); 
	end_tree = tree.end();
	pattern = "(.*)\\([A-Za-z\\.]+[0-9]*:[0-9\\.]+,(.*)";
	while(regex_search(start_tree, end_tree, what, pattern)){
		// First part of the tree
		string first_part = what[1];
		// Last part of the tree
		string last_part = what[2];
		
		// We must remove a balanced parenthesis in the last part. 
		// Scan from left to right
		int balanced = 0;
		for(int i = 0;i < last_part.length();++i){                                                     
			string letter = last_part.substr(i,1);
			
			if(letter == "("){
				++balanced;
				
			}
			else if(letter == ")"){
				if(balanced == 0){
					// Remove the encountered letter.
					tree = first_part + last_part.substr(0,i) + last_part.substr(i+1);
					break;
				}
				--balanced;	
			}
		}
		pattern2 = "(.+:)([0-9\\.]+):([0-9\\.]+)(.+)";
		std::string::const_iterator start_tree2, end_tree2;
		
		start_tree2 = tree.begin();
		end_tree2 = tree.end();
		
		if(regex_search(start_tree2,end_tree2,what2,pattern2)){
			string first;
			string second;
			first = what2[2];
			second = what2[3];
			double sum = atof(first.c_str()) + atof(second.c_str());
			
			stringstream out;
			out << sum;
			tree = what2[1] + out.str() + what2[4];
			out.clear();
			
		}
		start_tree = tree.begin(); 
		end_tree = tree.end();
	}
	
	start_tree = tree.begin(); 
	end_tree = tree.end();
	
	pattern = "(.*?):[0-9\\.]+$";
	if(regex_search(start_tree,end_tree,what,pattern)){
		tree = what[1];
	}
	
	// Here we populate the nodes with IDs.
	int parent = gen.getGenomeSize();
	int children_size = parent;
	start_tree = tree.begin(); 
	end_tree = tree.end();
	pattern = "\\(([0-9]*):([0-9\\.]*),([0-9]*):([0-9\\.]*)\\)";
	while(regex_search(start_tree, end_tree, what, pattern)) { 
		
		string child_1 = what[1];
		string child_2 = what[3];
		string distance_1 = what[2];
		string distance_2 = what[4];
		
		nodes.createNode();
		int node_ID = nodes.getNodeListSize()-1;
		nodes.setChildAsParent(node_ID);
	
		nodes.setChild1(node_ID,atoi(child_1.c_str()));
		nodes.setChild2(node_ID,atoi(child_2.c_str()));
		
		nodes.setDistance1(node_ID,d_scale * atof(distance_1.c_str()));
		nodes.setDistance2(node_ID,d_scale * atof(distance_2.c_str()));
		
		pattern2 = "(.*?)\\(([0-9:\\.]*),([0-9:\\.]*)\\)(.*)";
		stringstream out;
		out << parent;
		//tree = regex_replace(tree,pattern2,out.str());
		if(regex_search(start_tree,end_tree,what2,pattern2)){

			tree = what2[1] + out.str() + what2[4];	
		}
		out.clear();
		++parent;
				
		start_tree = tree.begin();
		end_tree = tree.end();
	}
	// Handle unrooted trees
	start_tree = tree.begin(); 
	end_tree = tree.end();
	pattern = "^\\(([0-9]*):([0-9\\.]*),([0-9]*):([0-9\\.]*,([0-9]*):([0-9\\.]*))\\)$";
	if(regex_search(start_tree, end_tree, what, pattern)) { 
		// Convert (a,b,c) to ((a,b),c)

		string child_1 = what[1];
		string child_2 = what[3];
		string child_3 = what[5];
		string distance_1 = what[2];
		string distance_2 = what[4];
		string distance_3 = what[6];
	
		nodes.createNode();
		int node_ID = nodes.getNodeListSize()-1;
		nodes.setChildAsParent(node_ID);
	
		nodes.setChild1(node_ID,atoi(child_1.c_str()));
		nodes.setChild2(node_ID,atoi(child_2.c_str()));
		
		nodes.setDistance1(node_ID,d_scale * atof(distance_1.c_str()));
		nodes.setDistance2(node_ID,d_scale * atof(distance_2.c_str()));

		++parent;
		nodes.createNode();
		node_ID = nodes.getNodeListSize()-1;
		nodes.setChildAsParent(node_ID);
	
		nodes.setChild1(node_ID,node_ID - 1);
		nodes.setChild2(node_ID,atoi(child_3.c_str()));
		
		nodes.setDistance1(node_ID,0);
		nodes.setDistance2(node_ID,d_scale * atof(distance_3.c_str()));
		stringstream out;
		out << parent;
		tree = out.str();
		out.clear();

		
		++parent;
		
	}
	// Verify that tree has been correctly parsed
	if(parent == 0){
		cout << "Failed to parse the tree. No species could be found?" << endl;
		return 0;
	}
	pattern = "[^0-9]";
	start_tree = tree.begin(); 
	end_tree = tree.end();
	if(regex_search(start_tree,end_tree,what,pattern)){
		cout << "Tree not correctly parsed. Correct parenthesis?" << endl;
		return 0;
	}

	// We should make sure all child are equal in length; Skip this for now.
	if(stop == 0){ 
		stop = gen.getSequenceByID(0).getSequence().length();
	}	
	
	
	//###############################################################################
	// Populate distance distribution
	// By default, the distances are scaled by the distance flag.
	// Additionally, if the distance distribution is different at each position,
	// have a tab delimited file with the position and the scaling factor.
	// Example:
	// 1	0.5
	// 2	0.8
	//###############################################################################
	double distance_dist[gen.getSequenceByID(0).getSequence().length()];
	memset(distance_dist,'\0',sizeof(double)*(gen.getSequenceByID(0).getSequence().length()));
	pattern = "^([0-9]+)\\t([0-9\\.]+)";
	if(distrib_file == ""){
		for(int i = 0;i < gen.getSequenceByID(0).getSequence().length();++i){
			distance_dist[i] = 1;	
		}
	}
	else{
		ifstream file;
		file.open(distrib_file.c_str(),ifstream::in);
		if(!file) return 0;
		string line;
		while(!file.fail()){
			getline(file,line);
			start_line = line.begin(); 
			end_line = line.end();
			if(regex_search(start_line, end_line, what, pattern)) {
				string pos_dist = what[1];
				string pos_val = what[2];
				distance_dist[atoi(pos_dist.c_str())-1] = atof(pos_val.c_str());
			}
			
		}
		file.close();
		for(int i = 0;i < gen.getSequenceByID(0).getSequence().length();++i){
			if(distance_dist[i] == 0){
				distance_dist[i] = 1;	
			}
		}
	}	
	
	//######################################################################
	// Populate substitution matrix for the substitution model
	// We will hardcode the matrix instead.
	//######################################################################
	double * sum_letters;
	sum_letters = new double[26];
	if(wag == 0){
		lambda[(int)('L' - 'A')]  = -0.262352069541849;
		lambda[(int)('S' - 'A')]  = -0.254867733800273;
		lambda[(int)('K' - 'A')]  = -0.225853382091412;
		lambda[(int)('I' - 'A')]  = -0.212332452414312;
		lambda[(int)('E' - 'A')]  = -0.188558928817862;
		lambda[(int)('N' - 'A')]  = -0.167140155639894;
		lambda[(int)('T' - 'A')]  = -0.14395950834156;
		lambda[(int)('D' - 'A')]  = -0.126531662282891;
		lambda[(int)('V' - 'A')]  = -0.124281165343963;
		lambda[(int)('A' - 'A')]  = -0.118051382908287;
		lambda[(int)('G' - 'A')]  = -0.112887472255798;
		lambda[(int)('R' - 'A')]  = -0.106768506411763;
		lambda[(int)('F' - 'A')]  = -0.0804401241329973;
		lambda[(int)('P' - 'A')]  = -0.0768501808813535;
		lambda[(int)('Q' - 'A')]  = -0.0748240469356219;
		lambda[(int)('Y' - 'A')]  = -0.0555950811873711;
		lambda[(int)('H' - 'A')]  = -0.0423826387755612;
		lambda[(int)('M' - 'A')]  = -0.0392766891199065;
		lambda[(int)('C' - 'A')]  = -0.0100967308691275;
		lambda[(int)('W' - 'A')]  = 0;
	
		P[0][0] = -0.049681237;	P[0][1] = -0.002415121;	P[0][2] = -0.017779188;	P[0][3] = -0.026492808;	P[0][4] = 0.003190431;	P[0][5] = -0.065180382;	P[0][6] = 0.22015554;	P[0][7] = 0.73416107;	P[0][8] = 0.29055798;	P[0][9] = -0.18070275;	P[0][10] = -0.12485798;	P[0][11] = 0.28014971;	P[0][12] = -0.1813374;	P[0][13] = -0.012401368;	P[0][14] = -0.046902305;	P[0][15] = -0.035150013;	P[0][16] = -0.32385684;	P[0][17] = 0.03378741;	P[0][18] = -0.033768978;	P[0][19] = 0.2236068;
		P[1][0] = 0.81716129;	P[1][1] = 0.061249572;	P[1][2] = 0.15587172;	P[1][3] = 0.31333291;	P[1][4] = 0.066562022;	P[1][5] = 0.15059126;	P[1][6] = 0.23673576;	P[1][7] = 0.040072747;	P[1][8] = -0.10097594;	P[1][9] = -0.031716243;	P[1][10] = 0.032901684;	P[1][11] = -0.070946219;	P[1][12] = 0.065944418;	P[1][13] = -0.036255272;	P[1][14] = 0.19477143;	P[1][15] = 0.079643091;	P[1][16] = -0.001441757;	P[1][17] = -0.037550607;	P[1][18] = -0.050915707;	P[1][19] = 0.2236068;
		P[2][0] = 0.10517411;	P[2][1] = 0.00579841;	P[2][2] = -0.15827075;	P[2][3] = -0.41563643;	P[2][4] = 0.59004463;	P[2][5] = 0.045540384;	P[2][6] = 0.002593839;	P[2][7] = 0.086828695;	P[2][8] = 0.026265838;	P[2][9] = 0.15779057;	P[2][10] = -0.20092364;	P[2][11] = -0.39370815;	P[2][12] = -0.24896154;	P[2][13] = 0.028638242;	P[2][14] = -0.15548169;	P[2][15] = -0.07368193;	P[2][16] = 0.20682824;	P[2][17] = -0.15559485;	P[2][18] = -0.054511853;	P[2][19] = 0.2236068;
		P[3][0] = 0.058081544;	P[3][1] = -0.59299062;	P[3][2] = -0.25514888;	P[3][3] = 0.19477615;	P[3][4] = 0.063859591;	P[3][5] = 0.18749096;	P[3][6] = -0.41081291;	P[3][7] = -0.00201392;	P[3][8] = 0.11120814;	P[3][9] = 0.12305416;	P[3][10] = 0.28544197;	P[3][11] = -0.071589204;	P[3][12] = 0.076413924;	P[3][13] = -0.017739701;	P[3][14] = -0.17796799;	P[3][15] = 0.067092985;	P[3][16] = -0.34091978;	P[3][17] = -0.089891867;	P[3][18] = -0.056580404;	P[3][19] = 0.2236068;
		P[4][0] = -0.047533081;	P[4][1] = -0.014890516;	P[4][2] = 0.15818656;	P[4][3] = 0.38193051;	P[4][4] = 0.19960266;	P[4][5] = -0.56531243;	P[4][6] = -0.12978641;	P[4][7] = -0.044030883;	P[4][8] = 0.15943821;	P[4][9] = 0.35145129;	P[4][10] = -0.12259817;	P[4][11] = 0.26434271;	P[4][12] = 0.070585802;	P[4][13] = 0.26961129;	P[4][14] = -0.000086778831;	P[4][15] = -0.05864428;	P[4][16] = 0.21638688;	P[4][17] = -0.18148197;	P[4][18] = -0.060874437;	P[4][19] = 0.2236068;
		P[5][0] = -0.44696064;	P[5][1] = -0.074629599;	P[5][2] = 0.40818135;	P[5][3] = 0.22989943;	P[5][4] = 0.1947956;	P[5][5] = 0.53009713;	P[5][6] = 0.15888675;	P[5][7] = 0.15108131;	P[5][8] = -0.30671582;	P[5][9] = 0.070081118;	P[5][10] = 0.026053068;	P[5][11] = 0.013175221;	P[5][12] = 0.11931146;	P[5][13] = 0.1047864;	P[5][14] = 0.048731598;	P[5][15] = -0.011323853;	P[5][16] = 0.13459322;	P[5][17] = -0.095339243;	P[5][18] = -0.0583983;	P[5][19] = 0.2236068;
		P[6][0] = -0.25165371;	P[6][1] = 0.28614434;	P[6][2] = -0.61615283;	P[6][3] = 0.27315449;	P[6][4] = 0.007468631;	P[6][5] = -0.11026103;	P[6][6] = 0.43010446;	P[6][7] = -0.075518941;	P[6][8] = -0.09575316;	P[6][9] = 0.070162308;	P[6][10] = 0.19633815;	P[6][11] = -0.2000809;	P[6][12] = 0.11319097;	P[6][13] = -0.038537922;	P[6][14] = 0.080977783;	P[6][15] = 0.074447331;	P[6][16] = -0.11512003;	P[6][17] = -0.10933631;	P[6][18] = -0.057889781;	P[6][19] = 0.2236068;
		P[7][0] = 0.14269558;	P[7][1] = 0.028835636;	P[7][2] = -0.29242293;	P[7][3] = -0.38034106;	P[7][4] = -0.42496808;	P[7][5] = 0.19846483;	P[7][6] = -0.09024041;	P[7][7] = 0.14867582;	P[7][8] = -0.19645181;	P[7][9] = 0.2957769;	P[7][10] = -0.10008594;	P[7][11] = 0.28964442;	P[7][12] = 0.23053356;	P[7][13] = 0.31858099;	P[7][14] = 0.070353596;	P[7][15] = -0.030264684;	P[7][16] = 0.20105761;	P[7][17] = -0.15388422;	P[7][18] = -0.06167617;	P[7][19] = 0.2236068;
		P[8][0] = -0.00671921;	P[8][1] = 0.64479343;	P[8][2] = 0.29093514;	P[8][3] = -0.15579176;	P[8][4] = -0.015026826;	P[8][5] = 0.044523791;	P[8][6] = -0.34448917;	P[8][7] = -0.001989732;	P[8][8] = 0.150836;	P[8][9] = 0.1265477;	P[8][10] = 0.34671352;	P[8][11] = -0.098619711;	P[8][12] = 0.10119853;	P[8][13] = -0.006939004;	P[8][14] = -0.11954331;	P[8][15] = 0.070286963;	P[8][16] = -0.30566775;	P[8][17] = -0.097616607;	P[8][18] = -0.056684905;	P[8][19] = 0.2236068;
		P[9][0] = -0.056616309;	P[9][1] = -0.37343765;	P[9][2] = 0.34804428;	P[9][3] = -0.42188597;	P[9][4] = -0.16969608;	P[9][5] = -0.35211762;	P[9][6] = 0.39387541;	P[9][7] = -0.1017408;	P[9][8] = 0.010508699;	P[9][9] = 0.03547992;	P[9][10] = 0.26407098;	P[9][11] = -0.19780946;	P[9][12] = 0.15314603;	P[9][13] = -0.021309736;	P[9][14] = 0.1374577;	P[9][15] = 0.08232548;	P[9][16] = -0.11543601;	P[9][17] = -0.11560482;	P[9][18] = -0.058673043;	P[9][19] = 0.2236068;
		P[10][0] = -0.058954396;	P[10][1] = 0.010706691;	P[10][2] = -0.064696527;	P[10][3] = -0.021783595;	P[10][4] = 0.05523449;	P[10][5] = 0.027676821;	P[10][6] = -0.15996143;	P[10][7] = -0.13748663;	P[10][8] = 0.26397134;	P[10][9] = -0.7291697;	P[10][10] = -0.11893473;	P[10][11] = -0.092767701;	P[10][12] = 0.29056045;	P[10][13] = 0.33461851;	P[10][14] = 0.16088686;	P[10][15] = 0.053018954;	P[10][16] = 0.13845924;	P[10][17] = -0.14782752;	P[10][18] = -0.056634067;	P[10][19] = 0.2236068;
		P[11][0] = -0.052229604;	P[11][1] = -0.006448166;	P[11][2] = 0.097640225;	P[11][3] = 0.24537214;	P[11][4] = -0.57456348;	P[11][5] = -0.072144571;	P[11][6] = -0.13903561;	P[11][7] = 0.14098981;	P[11][8] = -0.024022526;	P[11][9] = -0.049717429;	P[11][10] = -0.16132325;	P[11][11] = -0.49792519;	P[11][12] = -0.33609857;	P[11][13] = -0.070869891;	P[11][14] = -0.19068849;	P[11][15] = -0.074658126;	P[11][16] = 0.22846803;	P[11][17] = -0.1090617;	P[11][18] = -0.043999086;	P[11][19] = 0.2236068;
		P[12][0] = -0.019895157;	P[12][1] = 0.001037422;	P[12][2] = -0.002052849;	P[12][3] = -0.010189739;	P[12][4] = -0.002269018;	P[12][5] = -0.00604177;	P[12][6] = -0.040261502;	P[12][7] = -0.20174552;	P[12][8] = -0.14090331;	P[12][9] = 0.011250757;	P[12][10] = 0.044979776;	P[12][11] = -0.046092584;	P[12][12] = -0.37647547;	P[12][13] = 0.45009843;	P[12][14] = 0.29157229;	P[12][15] = -0.3754672;	P[12][16] = -0.26779468;	P[12][17] = 0.48930639;	P[12][18] = -0.039446127;	P[12][19] = 0.2236068;
		P[13][0] = -0.12562282;	P[13][1] = -0.002587705;	P[13][2] = -0.030032023;	P[13][3] = -0.051698466;	P[13][4] = 0.01066163;	P[13][5] = -0.001360316;	P[13][6] = -0.27446871;	P[13][7] = -0.058605507;	P[13][8] = -0.00208106;	P[13][9] = 0.071722469;	P[13][10] = -0.26755618;	P[13][11] = 0.071077241;	P[13][12] = -0.081213172;	P[13][13] = -0.54047866;	P[13][14] = 0.67396246;	P[13][15] = 0.060779576;	P[13][16] = -0.057459336;	P[13][17] = -0.082476431;	P[13][18] = -0.05684524;	P[13][19] = 0.2236068;
		P[14][0] = -0.000118842;	P[14][1] = 0.000522634;	P[14][2] = -0.010300904;	P[14][3] = -0.03107875;	P[14][4] = -0.089795219;	P[14][5] = 0.27649482;	P[14][6] = 0.22743903;	P[14][7] = -0.38928252;	P[14][8] = 0.44373145;	P[14][9] = -0.01490388;	P[14][10] = 0.21755774;	P[14][11] = 0.36884631;	P[14][12] = -0.35675943;	P[14][13] = -0.17926459;	P[14][14] = -0.13555852;	P[14][15] = -0.15474384;	P[14][16] = 0.25156734;	P[14][17] = -0.11299245;	P[14][18] = -0.053467688;	P[14][19] = 0.2236068;
		P[15][0] = -0.004703125;	P[15][1] = -0.00014131;	P[15][2] = -0.000049959719;	P[15][3] = 0.002107193;	P[15][4] = -0.015299426;	P[15][5] = 0.028701239;	P[15][6] = 0.041402345;	P[15][7] = 0.004054481;	P[15][8] = 0.22430225;	P[15][9] = 0.10998942;	P[15][10] = -0.16526757;	P[15][11] = -0.11546305;	P[15][12] = 0.52939247;	P[15][13] = -0.28923863;	P[15][14] = -0.18324817;	P[15][15] = -0.34477172;	P[15][16] = 0.13100243;	P[15][17] = 0.55575385;	P[15][18] = -0.041760599;	P[15][19] = 0.2236068;
		P[16][0] = 0.038982232;	P[16][1] = 0.007239687;	P[16][2] = -0.060140373;	P[16][3] = -0.062028214;	P[16][4] = 0.11571894;	P[16][5] = -0.27078197;	P[16][6] = -0.18122503;	P[16][7] = 0.11423333;	P[16][8] = -0.52948374;	P[16][9] = -0.35321224;	P[16][10] = 0.36961101;	P[16][11] = 0.25098882;	P[16][12] = -0.053141185;	P[16][13] = -0.27033524;	P[16][14] = -0.16669016;	P[16][15] = -0.18630698;	P[16][16] = 0.24147237;	P[16][17] = 0.064615376;	P[16][18] = -0.052473005;	P[16][19] = 0.2236068;
		P[17][0] = 0.009115137;	P[17][1] = 0.025056105;	P[17][2] = 0.06398273;	P[17][3] = -0.029439419;	P[17][4] = -0.028172138;	P[17][5] = -0.027853075;	P[17][6] = 0.092631083;	P[17][7] = -0.37531455;	P[17][8] = -0.29037999;	P[17][9] = -0.10286728;	P[17][10] = -0.53120924;	P[17][11] = 0.18304029;	P[17][12] = 0.009323667;	P[17][13] = -0.080292469;	P[17][14] = -0.41227551;	P[17][15] = 0.079080378;	P[17][16] = -0.42820641;	P[17][17] = -0.11641769;	P[17][18] = -0.055281138;	P[17][19] = 0.2236068;
		P[18][0] = -0.048576489;	P[18][1] = -0.003272976;	P[18][2] = -0.015034294;	P[18][3] = -0.032018034;	P[18][4] = 0.008686522;	P[18][5] = -0.019220409;	P[18][6] = -0.030547168;	P[18][7] = -0.019603867;	P[18][8] = 0.012652466;	P[18][9] = 0.032529991;	P[18][10] = 0.00431469;	P[18][11] = 0.06268089;	P[18][12] = -0.14034989;	P[18][13] = 0.056782361;	P[18][14] = -0.075644002;	P[18][15] = 0.78980015;	P[18][16] = 0.2055114;	P[18][17] = 0.49418157;	P[18][18] = -0.023902416;	P[18][19] = 0.2236068;
		P[19][0] = -0.001945281;	P[19][1] = -0.000570264;	P[19][2] = -0.000760496;	P[19][3] = -0.002188577;	P[19][4] = 0.003965127;	P[19][5] = 0.000692331;	P[19][6] = -0.002995869;	P[19][7] = -0.012764388;	P[19][8] = -0.006705005;	P[19][9] = 0.006452907;	P[19][10] = 0.004774105;	P[19][11] = 0.001056551;	P[19][12] = 0.014735376;	P[19][13] = 0.000546254;	P[19][14] = 0.005373195;	P[19][15] = -0.011462277;	P[19][16] = 0.000555835;	P[19][17] = -0.032568334;	P[19][18] = 0.97378294;	P[19][19] = 0.2236068;

		P1[0][0] =  -0.049681237;	P1[0][1] =  0.81716129;	P1[0][2] =  0.10517411;	P1[0][3] =  0.058081544;	P1[0][4] =  -0.047533081;	P1[0][5] =  -0.44696064;	P1[0][6] =  -0.25165371;	P1[0][7] =  0.14269558;	P1[0][8] =  -0.00671921;	P1[0][9] =  -0.056616309;	P1[0][10] =  -0.058954396;	P1[0][11] =  -0.052229604;	P1[0][12] =  -0.019895157;	P1[0][13] =  -0.12562282;	P1[0][14] =  -0.000118842;	P1[0][15] =  -0.004703125;	P1[0][16] =  0.038982232;	P1[0][17] =  0.009115137;	P1[0][18] =  -0.048576489;	P1[0][19] =  -0.001945281;
		P1[1][0] =  -0.002415121;	P1[1][1] =  0.061249572;	P1[1][2] =  0.00579841;	P1[1][3] =  -0.59299062;	P1[1][4] =  -0.014890516;	P1[1][5] =  -0.074629599;	P1[1][6] =  0.28614434;	P1[1][7] =  0.028835636;	P1[1][8] =  0.64479343;	P1[1][9] =  -0.37343765;	P1[1][10] =  0.010706691;	P1[1][11] =  -0.006448166;	P1[1][12] =  0.001037422;	P1[1][13] =  -0.002587705;	P1[1][14] =  0.000522634;	P1[1][15] =  -0.00014131;	P1[1][16] =  0.007239687;	P1[1][17] =  0.025056105;	P1[1][18] =  -0.003272976;	P1[1][19] =  -0.000570264;
		P1[2][0] =  -0.017779188;	P1[2][1] =  0.15587172;	P1[2][2] =  -0.15827075;	P1[2][3] =  -0.25514888;	P1[2][4] =  0.15818656;	P1[2][5] =  0.40818135;	P1[2][6] =  -0.61615283;	P1[2][7] =  -0.29242293;	P1[2][8] =  0.29093514;	P1[2][9] =  0.34804428;	P1[2][10] =  -0.064696527;	P1[2][11] =  0.097640225;	P1[2][12] =  -0.002052849;	P1[2][13] =  -0.030032023;	P1[2][14] =  -0.010300904;	P1[2][15] =  -0.000049959719;	P1[2][16] =  -0.060140373;	P1[2][17] =  0.06398273;	P1[2][18] =  -0.015034294;	P1[2][19] =  -0.000760496;
		P1[3][0] =  -0.026492808;	P1[3][1] =  0.31333291;	P1[3][2] =  -0.41563643;	P1[3][3] =  0.19477615;	P1[3][4] =  0.38193051;	P1[3][5] =  0.22989943;	P1[3][6] =  0.27315449;	P1[3][7] =  -0.38034106;	P1[3][8] =  -0.15579176;	P1[3][9] =  -0.42188597;	P1[3][10] =  -0.021783595;	P1[3][11] =  0.24537214;	P1[3][12] =  -0.010189739;	P1[3][13] =  -0.051698466;	P1[3][14] =  -0.03107875;	P1[3][15] =  0.002107193;	P1[3][16] =  -0.062028214;	P1[3][17] =  -0.029439419;	P1[3][18] =  -0.032018034;	P1[3][19] =  -0.002188577;
		P1[4][0] =  0.003190431;	P1[4][1] =  0.066562022;	P1[4][2] =  0.59004463;	P1[4][3] =  0.063859591;	P1[4][4] =  0.19960266;	P1[4][5] =  0.1947956;	P1[4][6] =  0.007468631;	P1[4][7] =  -0.42496808;	P1[4][8] =  -0.015026826;	P1[4][9] =  -0.16969608;	P1[4][10] =  0.05523449;	P1[4][11] =  -0.57456348;	P1[4][12] =  -0.002269018;	P1[4][13] =  0.01066163;	P1[4][14] =  -0.089795219;	P1[4][15] =  -0.015299426;	P1[4][16] =  0.11571894;	P1[4][17] =  -0.028172138;	P1[4][18] =  0.008686522;	P1[4][19] =  0.003965127;
		P1[5][0] =  -0.065180382;	P1[5][1] =  0.15059126;	P1[5][2] =  0.045540384;	P1[5][3] =  0.18749096;	P1[5][4] =  -0.56531243;	P1[5][5] =  0.53009713;	P1[5][6] =  -0.11026103;	P1[5][7] =  0.19846483;	P1[5][8] =  0.044523791;	P1[5][9] =  -0.35211762;	P1[5][10] =  0.027676821;	P1[5][11] =  -0.072144571;	P1[5][12] =  -0.00604177;	P1[5][13] =  -0.001360316;	P1[5][14] =  0.27649482;	P1[5][15] =  0.028701239;	P1[5][16] =  -0.27078197;	P1[5][17] =  -0.027853075;	P1[5][18] =  -0.019220409;	P1[5][19] =  0.000692331;
		P1[6][0] =  0.22015554;	P1[6][1] =  0.23673576;	P1[6][2] =  0.002593839;	P1[6][3] =  -0.41081291;	P1[6][4] =  -0.12978641;	P1[6][5] =  0.15888675;	P1[6][6] =  0.43010446;	P1[6][7] =  -0.09024041;	P1[6][8] =  -0.34448917;	P1[6][9] =  0.39387541;	P1[6][10] =  -0.15996143;	P1[6][11] =  -0.13903561;	P1[6][12] =  -0.040261502;	P1[6][13] =  -0.27446871;	P1[6][14] =  0.22743903;	P1[6][15] =  0.041402345;	P1[6][16] =  -0.18122503;	P1[6][17] =  0.092631083;	P1[6][18] =  -0.030547168;	P1[6][19] =  -0.002995869;
		P1[7][0] =  0.73416107;	P1[7][1] =  0.040072747;	P1[7][2] =  0.086828695;	P1[7][3] =  -0.00201392;	P1[7][4] =  -0.044030883;	P1[7][5] =  0.15108131;	P1[7][6] =  -0.075518941;	P1[7][7] =  0.14867582;	P1[7][8] =  -0.001989732;	P1[7][9] =  -0.1017408;	P1[7][10] =  -0.13748663;	P1[7][11] =  0.14098981;	P1[7][12] =  -0.20174552;	P1[7][13] =  -0.058605507;	P1[7][14] =  -0.38928252;	P1[7][15] =  0.004054481;	P1[7][16] =  0.11423333;	P1[7][17] =  -0.37531455;	P1[7][18] =  -0.019603867;	P1[7][19] =  -0.012764388;
		P1[8][0] =  0.29055798;	P1[8][1] =  -0.10097594;	P1[8][2] =  0.026265838;	P1[8][3] =  0.11120814;	P1[8][4] =  0.15943821;	P1[8][5] =  -0.30671582;	P1[8][6] =  -0.09575316;	P1[8][7] =  -0.19645181;	P1[8][8] =  0.150836;	P1[8][9] =  0.010508699;	P1[8][10] =  0.26397134;	P1[8][11] =  -0.024022526;	P1[8][12] =  -0.14090331;	P1[8][13] =  -0.00208106;	P1[8][14] =  0.44373145;	P1[8][15] =  0.22430225;	P1[8][16] =  -0.52948374;	P1[8][17] =  -0.29037999;	P1[8][18] =  0.012652466;	P1[8][19] =  -0.006705005;
		P1[9][0] =  -0.18070275;	P1[9][1] =  -0.031716243;	P1[9][2] =  0.15779057;	P1[9][3] =  0.12305416;	P1[9][4] =  0.35145129;	P1[9][5] =  0.070081118;	P1[9][6] =  0.070162308;	P1[9][7] =  0.2957769;	P1[9][8] =  0.1265477;	P1[9][9] =  0.03547992;	P1[9][10] =  -0.7291697;	P1[9][11] =  -0.049717429;	P1[9][12] =  0.011250757;	P1[9][13] =  0.071722469;	P1[9][14] =  -0.01490388;	P1[9][15] =  0.10998942;	P1[9][16] =  -0.35321224;	P1[9][17] =  -0.10286728;	P1[9][18] =  0.032529991;	P1[9][19] =  0.006452907;
		P1[10][0] =  -0.12485798;	P1[10][1] =  0.032901684;	P1[10][2] =  -0.20092364;	P1[10][3] =  0.28544197;	P1[10][4] =  -0.12259817;	P1[10][5] =  0.026053068;	P1[10][6] =  0.19633815;	P1[10][7] =  -0.10008594;	P1[10][8] =  0.34671352;	P1[10][9] =  0.26407098;	P1[10][10] =  -0.11893473;	P1[10][11] =  -0.16132325;	P1[10][12] =  0.044979776;	P1[10][13] =  -0.26755618;	P1[10][14] =  0.21755774;	P1[10][15] =  -0.16526757;	P1[10][16] =  0.36961101;	P1[10][17] =  -0.53120924;	P1[10][18] =  0.00431469;	P1[10][19] =  0.004774105;
		P1[11][0] =  0.28014971;	P1[11][1] =  -0.070946219;	P1[11][2] =  -0.39370815;	P1[11][3] =  -0.071589204;	P1[11][4] =  0.26434271;	P1[11][5] =  0.013175221;	P1[11][6] =  -0.2000809;	P1[11][7] =  0.28964442;	P1[11][8] =  -0.098619711;	P1[11][9] =  -0.19780946;	P1[11][10] =  -0.092767701;	P1[11][11] =  -0.49792519;	P1[11][12] =  -0.046092584;	P1[11][13] =  0.071077241;	P1[11][14] =  0.36884631;	P1[11][15] =  -0.11546305;	P1[11][16] =  0.25098882;	P1[11][17] =  0.18304029;	P1[11][18] =  0.06268089;	P1[11][19] =  0.001056551;
		P1[12][0] =  -0.1813374;	P1[12][1] =  0.065944418;	P1[12][2] =  -0.24896154;	P1[12][3] =  0.076413924;	P1[12][4] =  0.070585802;	P1[12][5] =  0.11931146;	P1[12][6] =  0.11319097;	P1[12][7] =  0.23053356;	P1[12][8] =  0.10119853;	P1[12][9] =  0.15314603;	P1[12][10] =  0.29056045;	P1[12][11] =  -0.33609857;	P1[12][12] =  -0.37647547;	P1[12][13] =  -0.081213172;	P1[12][14] =  -0.35675943;	P1[12][15] =  0.52939247;	P1[12][16] =  -0.053141185;	P1[12][17] =  0.009323667;	P1[12][18] =  -0.14034989;	P1[12][19] =  0.014735376;
		P1[13][0] =  -0.012401368;	P1[13][1] =  -0.036255272;	P1[13][2] =  0.028638242;	P1[13][3] =  -0.017739701;	P1[13][4] =  0.26961129;	P1[13][5] =  0.1047864;	P1[13][6] =  -0.038537922;	P1[13][7] =  0.31858099;	P1[13][8] =  -0.006939004;	P1[13][9] =  -0.021309736;	P1[13][10] =  0.33461851;	P1[13][11] =  -0.070869891;	P1[13][12] =  0.45009843;	P1[13][13] =  -0.54047866;	P1[13][14] =  -0.17926459;	P1[13][15] =  -0.28923863;	P1[13][16] =  -0.27033524;	P1[13][17] =  -0.080292469;	P1[13][18] =  0.056782361;	P1[13][19] =  0.000546254;
		P1[14][0] =  -0.046902305;	P1[14][1] =  0.19477143;	P1[14][2] =  -0.15548169;	P1[14][3] =  -0.17796799;	P1[14][4] =  -0.000086778831;	P1[14][5] =  0.048731598;	P1[14][6] =  0.080977783;	P1[14][7] =  0.070353596;	P1[14][8] =  -0.11954331;	P1[14][9] =  0.1374577;	P1[14][10] =  0.16088686;	P1[14][11] =  -0.19068849;	P1[14][12] =  0.29157229;	P1[14][13] =  0.67396246;	P1[14][14] =  -0.13555852;	P1[14][15] =  -0.18324817;	P1[14][16] =  -0.16669016;	P1[14][17] =  -0.41227551;	P1[14][18] =  -0.075644002;	P1[14][19] =  0.005373195;
		P1[15][0] =  -0.035150013;	P1[15][1] =  0.079643091;	P1[15][2] =  -0.07368193;	P1[15][3] =  0.067092985;	P1[15][4] =  -0.05864428;	P1[15][5] =  -0.011323853;	P1[15][6] =  0.074447331;	P1[15][7] =  -0.030264684;	P1[15][8] =  0.070286963;	P1[15][9] =  0.08232548;	P1[15][10] =  0.053018954;	P1[15][11] =  -0.074658126;	P1[15][12] =  -0.3754672;	P1[15][13] =  0.060779576;	P1[15][14] =  -0.15474384;	P1[15][15] =  -0.34477172;	P1[15][16] =  -0.18630698;	P1[15][17] =  0.079080378;	P1[15][18] =  0.78980015;	P1[15][19] =  -0.011462277;
		P1[16][0] =  -0.32385684;	P1[16][1] =  -0.001441757;	P1[16][2] =  0.20682824;	P1[16][3] =  -0.34091978;	P1[16][4] =  0.21638688;	P1[16][5] =  0.13459322;	P1[16][6] =  -0.11512003;	P1[16][7] =  0.20105761;	P1[16][8] =  -0.30566775;	P1[16][9] =  -0.11543601;	P1[16][10] =  0.13845924;	P1[16][11] =  0.22846803;	P1[16][12] =  -0.26779468;	P1[16][13] =  -0.057459336;	P1[16][14] =  0.25156734;	P1[16][15] =  0.13100243;	P1[16][16] =  0.24147237;	P1[16][17] =  -0.42820641;	P1[16][18] =  0.2055114;	P1[16][19] =  0.000555835;
		P1[17][0] =  0.03378741;	P1[17][1] =  -0.037550607;	P1[17][2] =  -0.15559485;	P1[17][3] =  -0.089891867;	P1[17][4] =  -0.18148197;	P1[17][5] =  -0.095339243;	P1[17][6] =  -0.10933631;	P1[17][7] =  -0.15388422;	P1[17][8] =  -0.097616607;	P1[17][9] =  -0.11560482;	P1[17][10] =  -0.14782752;	P1[17][11] =  -0.1090617;	P1[17][12] =  0.48930639;	P1[17][13] =  -0.082476431;	P1[17][14] =  -0.11299245;	P1[17][15] =  0.55575385;	P1[17][16] =  0.064615376;	P1[17][17] =  -0.11641769;	P1[17][18] =  0.49418157;	P1[17][19] =  -0.032568334;
		P1[18][0] =  -0.033768978;	P1[18][1] =  -0.050915707;	P1[18][2] =  -0.054511853;	P1[18][3] =  -0.056580404;	P1[18][4] =  -0.060874437;	P1[18][5] =  -0.0583983;	P1[18][6] =  -0.057889781;	P1[18][7] =  -0.06167617;	P1[18][8] =  -0.056684905;	P1[18][9] =  -0.058673043;	P1[18][10] =  -0.056634067;	P1[18][11] =  -0.043999086;	P1[18][12] =  -0.039446127;	P1[18][13] =  -0.05684524;	P1[18][14] =  -0.053467688;	P1[18][15] =  -0.041760599;	P1[18][16] =  -0.052473005;	P1[18][17] =  -0.055281138;	P1[18][18] =  -0.023902416;	P1[18][19] =  0.97378294;
		P1[19][0] =  0.2236068;	P1[19][1] =  0.2236068;	P1[19][2] =  0.2236068;	P1[19][3] =  0.2236068;	P1[19][4] =  0.2236068;	P1[19][5] =  0.2236068;	P1[19][6] =  0.2236068;	P1[19][7] =  0.2236068;	P1[19][8] =  0.2236068;	P1[19][9] =  0.2236068;	P1[19][10] =  0.2236068;	P1[19][11] =  0.2236068;	P1[19][12] =  0.2236068;	P1[19][13] =  0.2236068;	P1[19][14] =  0.2236068;	P1[19][15] =  0.2236068;	P1[19][16] =  0.2236068;	P1[19][17] =  0.2236068;	P1[19][18] =  0.2236068;	P1[19][19] =  0.2236068;

		sum_letters[0]  = 0.0963342053386846;
		sum_letters[1]  = 0.0882158107506429;
		sum_letters[2]  = 0.0732072974102184;
		sum_letters[3]  = 0.0637063066162858;
		sum_letters[4]  = 0.0660862696594149;
		sum_letters[5]  = 0.0602300973875848;
		sum_letters[6]  = 0.0577678301469987;
		sum_letters[7]  = 0.0608816722841187;
		sum_letters[8]  = 0.0562852114789207;
		sum_letters[9]  = 0.0558982674394878;
		sum_letters[10]  = 0.0505473514909348;
		sum_letters[11]  = 0.0444825609408284;
		sum_letters[12]  = 0.0441144782799794;
		sum_letters[13]  = 0.0442087851728988;
		sum_letters[14]  = 0.0403324860954489;
		sum_letters[15]  = 0.0326479032148362;
		sum_letters[16]  = 0.0216785826759952;
		sum_letters[17]  = 0.0207315128483139;
		sum_letters[18]  = 0.0122581814087392;
		sum_letters[19]  = 0.0103851893596676;
	}
	else{
		lambda[(int)('L' - 'A')]  = -0.189812092944174;
		lambda[(int)('S' - 'A')]  = -3.00000044953125E-07;
		lambda[(int)('K' - 'A')]  = -0.161670618826804;
		lambda[(int)('I' - 'A')]  = -0.157214376490117;
		lambda[(int)('E' - 'A')]  = -0.152909355589079;
		lambda[(int)('N' - 'A')]  = -0.0253915748075592;
		lambda[(int)('T' - 'A')]  = -0.137019870293421;
		lambda[(int)('D' - 'A')]  = -0.0340912075902619;
		lambda[(int)('V' - 'A')]  = -0.0409734018150945;
		lambda[(int)('A' - 'A')]  = -0.130538401807701;
		lambda[(int)('G' - 'A')]  = -0.119324019862902;
		lambda[(int)('R' - 'A')]  = -0.0495939706353095;
		lambda[(int)('F' - 'A')]  = -0.0547153851720098;
		lambda[(int)('P' - 'A')]  = -0.0587445016006264;
		lambda[(int)('Q' - 'A')]  = -0.107462334042498;
		lambda[(int)('Y' - 'A')]  = -0.074066423786098;
		lambda[(int)('H' - 'A')]  = -0.0980855756547957;
		lambda[(int)('M' - 'A')]  = -0.0962850778630469;
		lambda[(int)('C' - 'A')]  = -0.0858769915572378;
		
		P[0][0] = 0.091054625;	P[0][1] = -0.22360951;	P[0][2] = 0.70545772;	P[0][3] = 0.23883865;	P[0][4] = 0.13368332;	P[0][5] = 0.059742572;	P[0][6] = -0.28980527;	P[0][7] = 0.18080128;	P[0][8] = 0.26339188;	P[0][9] = -0.025532082;	P[0][10] = 0.072476512;	P[0][11] = -0.14277194;	P[0][12] = 0.037873135;	P[0][13] = -0.031140868;	P[0][14] = 0.039731329;	P[0][15] = 0.059903092;	P[0][16] = 0.25149469;	P[0][17] = -0.082064798;	P[0][18] = -0.27492555;	P[0][19] = -0.010241735;
		P[1][0] = -0.50203557;	P[1][1] = -0.22360826;	P[1][2] = 0.28273666;	P[1][3] = -0.54840356;	P[1][4] = -0.28889045;	P[1][5] = 0.022198317;	P[1][6] = -0.029974738;	P[1][7] = -0.09897819;	P[1][8] = -0.046088582;	P[1][9] = -0.11480396;	P[1][10] = 0.10005415;	P[1][11] = -0.026084326;	P[1][12] = -0.013851001;	P[1][13] = 0.08621139;	P[1][14] = -0.042957423;	P[1][15] = -0.080001047;	P[1][16] = 0.15883087;	P[1][17] = -0.25619551;	P[1][18] = 0.29111561;	P[1][19] = -0.071479025;
		P[2][0] = -0.017283163;	P[2][1] = -0.22360515;	P[2][2] = 0.056612688;	P[2][3] = -0.1572796;	P[2][4] = 0.22563233;	P[2][5] = 0.058778297;	P[2][6] = 0.0037734757;	P[2][7] = -0.14655599;	P[2][8] = -0.066022924;	P[2][9] = 0.73775567;	P[2][10] = -0.35139616;	P[2][11] = 0.0011327131;	P[2][12] = 0.089835679;	P[2][13] = -0.14510671;	P[2][14] = 0.025292546;	P[2][15] = 0.1627435;	P[2][16] = 0.012136342;	P[2][17] = 0.040934955;	P[2][18] = 0.06142347;	P[2][19] = -0.33455887;
		P[3][0] = 0.087740566;	P[3][1] = -0.22360959;	P[3][2] = 0.16731999;	P[3][3] = -0.075789256;	P[3][4] = -0.099708085;	P[3][5] = 0.057789484;	P[3][6] = 0.67627245;	P[3][7] = 0.13906959;	P[3][8] = 0.36882638;	P[3][9] = -0.018971558;	P[3][10] = -0.12231073;	P[3][11] = -0.29040155;	P[3][12] = 0.13349877;	P[3][13] = -0.03418322;	P[3][14] = 0.10538742;	P[3][15] = -0.036252559;	P[3][16] = -0.24891102;	P[3][17] = 0.19198918;	P[3][18] = 0.12826956;	P[3][19] = 0.19037128;
		P[4][0] = -0.11527562;	P[4][1] = -0.22361229;	P[4][2] = 0.020839499;	P[4][3] = 0.39750661;	P[4][4] = -0.60082117;	P[4][5] = 0.065694439;	P[4][6] = 0.015931455;	P[4][7] = -0.19272853;	P[4][8] = -0.089954102;	P[4][9] = 0.18966166;	P[4][10] = 0.12248475;	P[4][11] = 0.037424251;	P[4][12] = 0.090412046;	P[4][13] = -0.092032984;	P[4][14] = -0.29125188;	P[4][15] = -0.31873233;	P[4][16] = -0.096089483;	P[4][17] = 0.19509561;	P[4][18] = -0.19908013;	P[4][19] = -0.158567;
		P[5][0] = 0.056280828;	P[5][1] = -0.2236092;	P[5][2] = -0.066768239;	P[5][3] = 0.22575809;	P[5][4] = -0.16704824;	P[5][5] = 0.055957086;	P[5][6] = -0.10093924;	P[5][7] = -0.17419701;	P[5][8] = -0.088212458;	P[5][9] = -0.21887226;	P[5][10] = -0.57601279;	P[5][11] = 0.13139937;	P[5][12] = 0.11968347;	P[5][13] = -0.052304864;	P[5][14] = 0.5153492;	P[5][15] = -0.19490983;	P[5][16] = 0.12052663;	P[5][17] = -0.23131157;	P[5][18] = 0.092473946;	P[5][19] = 0.13367521;
		P[6][0] = 0.025933786;	P[6][1] = -0.22360866;	P[6][2] = -0.13824085;	P[6][3] = 0.43823988;	P[6][4] = 0.31454072;	P[6][5] = 0.033740995;	P[6][6] = 0.20580393;	P[6][7] = -0.094393202;	P[6][8] = 0.031107812;	P[6][9] = 0.014379639;	P[6][10] = 0.26690077;	P[6][11] = -0.090428246;	P[6][12] = 0.020309647;	P[6][13] = 0.020872236;	P[6][14] = -0.20406646;	P[6][15] = -0.1157518;	P[6][16] = 0.19337873;	P[6][17] = -0.40284698;	P[6][18] = 0.49009384;	P[6][19] = -0.10224813;
		P[7][0] = 0.013606676;	P[7][1] = -0.22361062;	P[7][2] = -0.004995016;	P[7][3] = -0.22920964;	P[7][4] = 0.4481811;	P[7][5] = 0.069961469;	P[7][6] = 0.060847587;	P[7][7] = -0.22177809;	P[7][8] = -0.1260506;	P[7][9] = -0.077295497;	P[7][10] = 0.30861569;	P[7][11] = 0.1005807;	P[7][12] = 0.12413203;	P[7][13] = -0.04336982;	P[7][14] = 0.24634832;	P[7][15] = -0.56880075;	P[7][16] = -0.069512804;	P[7][17] = 0.17794241;	P[7][18] = -0.2683052;	P[7][19] = -0.037296206;
		P[8][0] = -0.19863482;	P[8][1] = -0.22360883;	P[8][2] = -0.22800517;	P[8][3] = 0.027314597;	P[8][4] = 0.10511422;	P[8][5] = 0.033630218;	P[8][6] = -0.59488338;	P[8][7] = 0.077098026;	P[8][8] = 0.29889248;	P[8][9] = 0.0064628331;	P[8][10] = -0.01914737;	P[8][11] = -0.27264036;	P[8][12] = 0.10426893;	P[8][13] = -0.0060224582;	P[8][14] = -0.040577111;	P[8][15] = -0.079907918;	P[8][16] = -0.41764962;	P[8][17] = 0.18068771;	P[8][18] = 0.26301386;	P[8][19] = 0.16292375;
		P[9][0] = 0.81006422;	P[9][1] = -0.22360703;	P[9][2] = -0.047940704;	P[9][3] = -0.33867696;	P[9][4] = -0.24786309;	P[9][5] = -0.010502185;	P[9][6] = -0.19260452;	P[9][7] = -0.095362209;	P[9][8] = 0.0055011579;	P[9][9] = -0.0086609097;	P[9][10] = 0.10024163;	P[9][11] = -0.077312896;	P[9][12] = -0.021612113;	P[9][13] = 0.057106166;	P[9][14] = -0.13057755;	P[9][15] = -0.053051199;	P[9][16] = 0.029966857;	P[9][17] = -0.085195542;	P[9][18] = 0.13780212;	P[9][19] = -0.046719634;
		P[10][0] = -0.028080581;	P[10][1] = -0.22360932;	P[10][2] = -0.012213759;	P[10][3] = 0.086038869;	P[10][4] = 0.054801308;	P[10][5] = 0.045041909;	P[10][6] = 0.038073095;	P[10][7] = -0.16850832;	P[10][8] = -0.19339667;	P[10][9] = 0.0076382216;	P[10][10] = -0.030815087;	P[10][11] = -0.09566834;	P[10][12] = 0.13266633;	P[10][13] = 0.86690294;	P[10][14] = -0.0073609646;	P[10][15] = 0.23459406;	P[10][16] = -0.048627841;	P[10][17] = 0.082929205;	P[10][18] = -0.14468318;	P[10][19] = 0.069550548;
		P[11][0] = 0.0012134753;	P[11][1] = -0.22360777;	P[11][2] = -0.023703424;	P[11][3] = 0.09427355;	P[11][4] = -0.082673148;	P[11][5] = 0.036629317;	P[11][6] = 0.0043631193;	P[11][7] = -0.10006255;	P[11][8] = -0.12651398;	P[11][9] = -0.27850291;	P[11][10] = 0.33205046;	P[11][11] = 0.033884399;	P[11][12] = 0.094577081;	P[11][13] = -0.22930682;	P[11][14] = 0.37835284;	P[11][15] = 0.53854561;	P[11][16] = -0.16521979;	P[11][17] = 0.17677159;	P[11][18] = 0.096943963;	P[11][19] = -0.39332143;
		P[12][0] = -0.016274645;	P[12][1] = -0.22360988;	P[12][2] = -0.1325802;	P[12][3] = -0.035525673;	P[12][4] = -0.021103564;	P[12][5] = 0.072097367;	P[12][6] = 0.054711969;	P[12][7] = 0.37345021;	P[12][8] = 0.24173567;	P[12][9] = 0.0061805669;	P[12][10] = -0.014972304;	P[12][11] = 0.39762628;	P[12][12] = -0.18511401;	P[12][13] = 0.14717597;	P[12][14] = 0.0083443791;	P[12][15] = -0.036615658;	P[12][16] = -0.40366501;	P[12][17] = -0.44739441;	P[12][18] = -0.25478508;	P[12][19] = -0.28393887;
		P[13][0] = -0.043331317;	P[13][1] = -0.22360965;	P[13][2] = -0.028801744;	P[13][3] = 0.04830041;	P[13][4] = 0.034746623;	P[13][5] = 0.059993952;	P[13][6] = 0.035027474;	P[13][7] = -0.20955872;	P[13][8] = -0.068930737;	P[13][9] = 0.0013572205;	P[13][10] = -0.040223088;	P[13][11] = -0.25383376;	P[13][12] = -0.8886104;	P[13][13] = -0.061830409;	P[13][14] = 0.067189507;	P[13][15] = 0.036291502;	P[13][16] = -0.029071681;	P[13][17] = 0.054341149;	P[13][18] = -0.098754421;	P[13][19] = 0.10664713;
		P[14][0] = -0.01707626;	P[14][1] = -0.22360958;	P[14][2] = 0.0052537764;	P[14][3] = -0.07965585;	P[14][4] = 0.23811778;	P[14][5] = 0.066159945;	P[14][6] = 0.05311656;	P[14][7] = -0.16046821;	P[14][8] = -0.074844139;	P[14][9] = -0.47456179;	P[14][10] = -0.38584913;	P[14][11] = 0.068146905;	P[14][12] = 0.098224246;	P[14][13] = -0.20926201;	P[14][14] = -0.59931783;	P[14][15] = 0.14736242;	P[14][16] = -0.049119874;	P[14][17] = 0.061068559;	P[14][18] = -0.15326971;	P[14][19] = -0.073394469;
		P[15][0] = 0.011026047;	P[15][1] = -0.22360971;	P[15][2] = 0.0036482368;	P[15][3] = 0.026420435;	P[15][4] = 0.024318876;	P[15][5] = 0.067788884;	P[15][6] = -0.013127929;	P[15][7] = 0.24589231;	P[15][8] = 0.12930228;	P[15][9] = -0.007384494;	P[15][10] = -0.022245346;	P[15][11] = 0.57609145;	P[15][12] = -0.18939991;	P[15][13] = 0.12654787;	P[15][14] = -0.041938123;	P[15][15] = -0.025821081;	P[15][16] = 0.31522837;	P[15][17] = 0.52674531;	P[15][18] = 0.31650149;	P[15][19] = 0.053574723;
		P[16][0] = -0.013280479;	P[16][1] = -0.22360981;	P[16][2] = 0.0087193268;	P[16][3] = -0.025648073;	P[16][4] = -0.011632353;	P[16][5] = 0.073091487;	P[16][6] = 0.012962549;	P[16][7] = -0.1378315;	P[16][8] = -0.096282269;	P[16][9] = 0.2186361;	P[16][10] = 0.23335763;	P[16][11] = 0.32198383;	P[16][12] = 0.10819713;	P[16][13] = -0.22189768;	P[16][14] = -0.051386707;	P[16][15] = 0.29577076;	P[16][16] = -0.10307492;	P[16][17] = -0.20170732;	P[16][18] = -0.098027687;	P[16][19] = 0.7051635;
		P[17][0] = -0.11840265;	P[17][1] = -0.22360938;	P[17][2] = -0.54112533;	P[17][3] = -0.12008783;	P[17][4] = -0.085248353;	P[17][5] = 0.058786293;	P[17][6] = 0.02764003;	P[17][7] = 0.11873405;	P[17][8] = 0.28847802;	P[17][9] = 0.015997304;	P[17][10] = 0.069209881;	P[17][11] = -0.22135912;	P[17][12] = 0.11831956;	P[17][13] = -0.067922872;	P[17][14] = 0.04040041;	P[17][15] = 0.11519895;	P[17][16] = 0.5497466;	P[17][17] = -0.011140309;	P[17][18] = -0.36060996;	P[17][19] = -0.00087578136;
		P[18][0] = -0.033452759;	P[18][1] = -0.22356246;	P[18][2] = 0.0023980888;	P[18][3] = 0.026509773;	P[18][4] = 0.020736643;	P[18][5] = -0.97083699;	P[18][6] = 0.02414669;	P[18][7] = 0.0058601905;	P[18][8] = 0.0082559612;	P[18][9] = 0.010497446;	P[18][10] = -0.01670019;	P[18][11] = 0.034782281;	P[18][12] = -0.0067271241;	P[18][13] = -0.015134643;	P[18][14] = 0.0025445502;	P[18][15] = -0.005672624;	P[18][16] = -0.003521271;	P[18][17] = 0.015217687;	P[18][18] = -0.046367243;	P[18][19] = 0.014812105;
		P[19][0] = 0.0061977486;	P[19][1] = -0.22360924;	P[19][2] = -0.028610834;	P[19][3] = 0.0010699833;	P[19][4] = 0.0051139717;	P[19][5] = 0.044264298;	P[19][6] = 0.0086660113;	P[19][7] = 0.65950168;	P[19][8] = -0.65919471;	P[19][9] = 0.016016146;	P[19][10] = -0.025709822;	P[19][11] = -0.23255934;	P[19][12] = 0.033317375;	P[19][13] = -0.0952961;	P[19][14] = -0.019512703;	P[19][15] = -0.074899145;	P[19][16] = 0.0031703553;	P[19][17] = 0.014112588;	P[19][18] = 0.021192079;	P[19][19] = 0.075915948;
		
		P1[0][0] = 0.091052947;	P1[0][1] = -0.50203833;	P1[0][2] = -0.017277838;	P1[0][3] = 0.087740304;	P1[0][4] = -0.11528022;	P1[0][5] = 0.056280223;	P1[0][6] = 0.025933455;	P1[0][7] = 0.01360864;	P1[0][8] = -0.19863443;	P1[0][9] = 0.81006313;	P1[0][10] = -0.028080363;	P1[0][11] = 0.0012121415;	P1[0][12] = -0.016274461;	P1[0][13] = -0.043331061;	P1[0][14] = -0.017075813;	P1[0][15] = 0.011026092;	P1[0][16] = -0.013280401;	P1[0][17] = -0.11840035;	P1[0][18] = -0.03344176;	P1[0][19] = 0.0061977822;
		P1[1][0] = -0.2236076;	P1[1][1] = -0.22360063;	P1[1][2] = -0.22360696;	P1[1][3] = -0.22360828;	P1[1][4] = -0.2236072;	P1[1][5] = -0.22360658;	P1[1][6] = -0.22360083;	P1[1][7] = -0.22360722;	P1[1][8] = -0.22360752;	P1[1][9] = -0.22360593;	P1[1][10] = -0.22360679;	P1[1][11] = -0.22360752;	P1[1][12] = -0.2236086;	P1[1][13] = -0.22360701;	P1[1][14] = -0.22360741;	P1[1][15] = -0.22360823;	P1[1][16] = -0.22360793;	P1[1][17] = -0.22360618;	P1[1][18] = -0.22360878;	P1[1][19] = -0.22360876;
		P1[2][0] = 0.70543711;	P1[2][1] = 0.28278259;	P1[2][2] = 0.056682896;	P1[2][3] = 0.16732076;	P1[2][4] = 0.020728881;	P1[2][5] = -0.066810565;	P1[2][6] = -0.1382724;	P1[2][7] = -0.0049305271;	P1[2][8] = -0.22799895;	P1[2][9] = -0.047922033;	P1[2][10] = -0.012219952;	P1[2][11] = -0.023729269;	P1[2][12] = -0.13257725;	P1[2][13] = -0.028804688;	P1[2][14] = 0.0052772337;	P1[2][15] = 0.0036468811;	P1[2][16] = 0.0087233218;	P1[2][17] = -0.54111554;	P1[2][18] = 0.0023924803;	P1[2][19] = -0.028610265;
		P1[3][0] = 0.23874568;	P1[3][1] = -0.54846086;	P1[3][2] = -0.15726536;	P1[3][3] = -0.07581826;	P1[3][4] = 0.39746364;	P1[3][5] = 0.22575718;	P1[3][6] = 0.43828016;	P1[3][7] = -0.22918148;	P1[3][8] = 0.027356669;	P1[3][9] = -0.33868852;	P1[3][10] = 0.086044174;	P1[3][11] = 0.094270009;	P1[3][12] = -0.035506623;	P1[3][13] = 0.048306904;	P1[3][14] = -0.079642271;	P1[3][15] = 0.026420875;	P1[3][16] = -0.025649642;	P1[3][17] = -0.1200289;	P1[3][18] = 0.026516857;	P1[3][19] = 0.0010745368;
		P1[4][0] = 0.13377431;	P1[4][1] = -0.28890377;	P1[4][2] = 0.22563252;	P1[4][3] = -0.099696861;	P1[4][4] = -0.6007933;	P1[4][5] = -0.16703957;	P1[4][6] = 0.31455888;	P1[4][7] = 0.44816431;	P1[4][8] = 0.10508517;	P1[4][9] = -0.2478863;	P1[4][10] = 0.054805799;	P1[4][11] = -0.082671051;	P1[4][12] = -0.021122172;	P1[4][13] = 0.034746558;	P1[4][14] = 0.23810911;	P1[4][15] = 0.024321902;	P1[4][16] = -0.011632656;	P1[4][17] = -0.085293473;	P1[4][18] = 0.02073991;	P1[4][19] = 0.0051109242;
		P1[5][0] = 0.05971272;	P1[5][1] = 0.022229596;	P1[5][2] = 0.058775924;	P1[5][3] = 0.057758784;	P1[5][4] = 0.06569062;	P1[5][5] = 0.055958238;	P1[5][6] = 0.033772145;	P1[5][7] = 0.069959937;	P1[5][8] = 0.033609307;	P1[5][9] = -0.010500854;	P1[5][10] = 0.045043367;	P1[5][11] = 0.03662013;	P1[5][12] = 0.072056019;	P1[5][13] = 0.059993926;	P1[5][14] = 0.066152844;	P1[5][15] = 0.067757635;	P1[5][16] = 0.073079786;	P1[5][17] = 0.058758747;	P1[5][18] = -0.97085072;	P1[5][19] = 0.044217921;
		P1[6][0] = -0.28980515;	P1[6][1] = -0.02997398;	P1[6][2] = 0.0037668537;	P1[6][3] = 0.67627156;	P1[6][4] = 0.015928323;	P1[6][5] = -0.10093981;	P1[6][6] = 0.20580713;	P1[6][7] = 0.060850379;	P1[6][8] = -0.59488316;	P1[6][9] = -0.19260365;	P1[6][10] = 0.038072939;	P1[6][11] = 0.0043653335;	P1[6][12] = 0.054712005;	P1[6][13] = 0.035027275;	P1[6][14] = 0.053120335;	P1[6][15] = -0.013127919;	P1[6][16] = 0.012961442;	P1[6][17] = 0.027639974;	P1[6][18] = 0.024145111;	P1[6][19] = 0.0086659092;
		P1[7][0] = 0.18079849;	P1[7][1] = -0.098979062;	P1[7][2] = -0.14655603;	P1[7][3] = 0.13906719;	P1[7][4] = -0.1927348;	P1[7][5] = -0.17420058;	P1[7][6] = -0.09439458;	P1[7][7] = -0.22178391;	P1[7][8] = 0.077097199;	P1[7][9] = -0.095360464;	P1[7][10] = -0.16851095;	P1[7][11] = -0.10006316;	P1[7][12] = 0.37344495;	P1[7][13] = -0.20956253;	P1[7][14] = -0.16047226;	P1[7][15] = 0.24588694;	P1[7][16] = -0.13783678;	P1[7][17] = 0.11873222;	P1[7][18] = 0.0059285985;	P1[7][19] = 0.65949886;
		P1[8][0] = 0.26339172;	P1[8][1] = -0.04608943;	P1[8][2] = -0.066014585;	P1[8][3] = 0.36882631;	P1[8][4] = -0.089961017;	P1[8][5] = -0.088213783;	P1[8][6] = 0.031107316;	P1[8][7] = -0.12605556;	P1[8][8] = 0.29889232;	P1[8][9] = 0.0055009393;	P1[8][10] = -0.19339867;	P1[8][11] = -0.12651054;	P1[8][12] = 0.2417334;	P1[8][13] = -0.068932722;	P1[8][14] = -0.074844764;	P1[8][15] = 0.12929945;	P1[8][16] = -0.096284049;	P1[8][17] = 0.28847872;	P1[8][18] = 0.0082696763;	P1[8][19] = -0.65919406;
		P1[9][0] = -0.025494345;	P1[9][1] = -0.11480157;	P1[9][2] = 0.73775397;	P1[9][3] = -0.018978851;	P1[9][4] = 0.18965979;	P1[9][5] = -0.21887689;	P1[9][6] = 0.01438189;	P1[9][7] = -0.077291224;	P1[9][8] = 0.0064506842;	P1[9][9] = -0.0086606862;	P1[9][10] = 0.0076379301;	P1[9][11] = -0.27850299;	P1[9][12] = 0.0061690873;	P1[9][13] = 0.0013556533;	P1[9][14] = -0.47456558;	P1[9][15] = -0.0073817388;	P1[9][16] = 0.21863721;	P1[9][17] = 0.01600382;	P1[9][18] = 0.010500823;	P1[9][19] = 0.016014412;
		P1[10][0] = 0.072451966;	P1[10][1] = 0.10005098;	P1[10][2] = -0.35138834;	P1[10][3] = -0.12230544;	P1[10][4] = 0.12248824;	P1[10][5] = -0.57602091;	P1[10][6] = 0.26690126;	P1[10][7] = 0.30861723;	P1[10][8] = -0.019132883;	P1[10][9] = 0.10024106;	P1[10][10] = -0.030814406;	P1[10][11] = 0.33204824;	P1[10][12] = -0.014962296;	P1[10][13] = -0.040222151;	P1[10][14] = -0.38585102;	P1[10][15] = -0.022248715;	P1[10][16] = 0.23336229;	P1[10][17] = 0.069195592;	P1[10][18] = -0.0167115;	P1[10][19] = -0.025708327;
		P1[11][0] = -0.14277283;	P1[11][1] = -0.026085099;	P1[11][2] = 0.0011250205;	P1[11][3] = -0.29040251;	P1[11][4] = 0.037430349;	P1[11][5] = 0.1314001;	P1[11][6] = -0.090430069;	P1[11][7] = 0.10058555;	P1[11][8] = -0.27264074;	P1[11][9] = -0.077311837;	P1[11][10] = -0.095664813;	P1[11][11] = 0.033881166;	P1[11][12] = 0.39762504;	P1[11][13] = -0.25383126;	P1[11][14] = 0.068147075;	P1[11][15] = 0.57609088;	P1[11][16] = 0.32198455;	P1[11][17] = -0.22135969;	P1[11][18] = 0.034790502;	P1[11][19] = -0.23255861;
		P1[12][0] = 0.03787369;	P1[12][1] = -0.013850912;	P1[12][2] = 0.089840537;	P1[12][3] = 0.13349994;	P1[12][4] = 0.090408418;	P1[12][5] = 0.11968276;	P1[12][6] = 0.020310491;	P1[12][7] = 0.12412861;	P1[12][8] = 0.10426997;	P1[12][9] = -0.02161225;	P1[12][10] = 0.13266302;	P1[12][11] = 0.094580019;	P1[12][12] = -0.1851142;	P1[12][13] = -0.88861057;	P1[12][14] = 0.09822445;	P1[12][15] = -0.18940069;	P1[12][16] = 0.10819656;	P1[12][17] = 0.11831955;	P1[12][18] = -0.0067257086;	P1[12][19] = 0.033316959;
		P1[13][0] = -0.031141747;	P1[13][1] = 0.086211533;	P1[13][2] = -0.14510971;	P1[13][3] = -0.034186972;	P1[13][4] = -0.092030502;	P1[13][5] = -0.052304121;	P1[13][6] = 0.020871572;	P1[13][7] = -0.043367165;	P1[13][8] = -0.0060265183;	P1[13][9] = 0.057105183;	P1[13][10] = 0.8669025;	P1[13][11] = -0.22930893;	P1[13][12] = 0.14717678;	P1[13][13] = -0.061827211;	P1[13][14] = -0.20926163;	P1[13][15] = 0.12654997;	P1[13][16] = -0.221896;	P1[13][17] = -0.067921803;	P1[13][18] = -0.015144733;	P1[13][19] = -0.095295184;
		P1[14][0] = 0.039746857;	P1[14][1] = -0.042956533;	P1[14][2] = 0.025296168;	P1[14][3] = 0.10538133;	P1[14][4] = -0.29125463;	P1[14][5] = 0.51535379;	P1[14][6] = -0.20406788;	P1[14][7] = 0.24634729;	P1[14][8] = -0.040590963;	P1[14][9] = -0.13057786;	P1[14][10] = -0.0073611857;	P1[14][11] = 0.37834416;	P1[14][12] = 0.0083336593;	P1[14][13] = 0.067189229;	P1[14][14] = -0.59931602;	P1[14][15] = -0.041932131;	P1[14][16] = -0.051389359;	P1[14][17] = 0.040418367;	P1[14][18] = 0.0025534551;	P1[14][19] = -0.019513026;
		P1[15][0] = 0.059897546;	P1[15][1] = -0.079993517;	P1[15][2] = 0.16275292;	P1[15][3] = -0.036238813;	P1[15][4] = -0.31873875;	P1[15][5] = -0.19490997;	P1[15][6] = -0.11573723;	P1[15][7] = -0.5688097;	P1[15][8] = -0.079888561;	P1[15][9] = -0.053045837;	P1[15][10] = 0.23458651;	P1[15][11] = 0.53855346;	P1[15][12] = -0.036615059;	P1[15][13] = 0.036284543;	P1[15][14] = 0.14735905;	P1[15][15] = -0.025820564;	P1[15][16] = 0.29576062;	P1[15][17] = 0.11517783;	P1[15][18] = -0.0056636663;	P1[15][19] = -0.074900046;
		P1[16][0] = 0.25149574;	P1[16][1] = 0.15880415;	P1[16][2] = 0.012162569;	P1[16][3] = -0.24889263;	P1[16][4] = -0.096092054;	P1[16][5] = 0.12053879;	P1[16][6] = 0.19333418;	P1[16][7] = -0.069481829;	P1[16][8] = -0.41764078;	P1[16][9] = 0.029951008;	P1[16][10] = -0.048620127;	P1[16][11] = -0.16523793;	P1[16][12] = -0.40368678;	P1[16][13] = -0.029061843;	P1[16][14] = -0.04913248;	P1[16][15] = 0.3152514;	P1[16][16] = -0.10308001;	P1[16][17] = 0.54975002;	P1[16][18] = -0.0035335925;	P1[16][19] = 0.0031737286;
		P1[17][0] = -0.082054143;	P1[17][1] = -0.25618577;	P1[17][2] = 0.04093299;	P1[17][3] = 0.19197339;	P1[17][4] = 0.19508986;	P1[17][5] = -0.23130485;	P1[17][6] = -0.40283624;	P1[17][7] = 0.17793879;	P1[17][8] = 0.18066551;	P1[17][9] = -0.08519085;	P1[17][10] = 0.082925067;	P1[17][11] = 0.17676293;	P1[17][12] = -0.44741708;	P1[17][13] = 0.054338998;	P1[17][14] = 0.061063675;	P1[17][15] = 0.52676335;	P1[17][16] = -0.20171275;	P1[17][17] = -0.011111369;	P1[17][18] = 0.015244925;	P1[17][19] = 0.014112459;
		P1[18][0] = -0.27492262;	P1[18][1] = 0.2911152;	P1[18][2] = 0.061431947;	P1[18][3] = 0.12826771;	P1[18][4] = -0.19907014;	P1[18][5] = 0.092451265;	P1[18][6] = 0.49009606;	P1[18][7] = -0.2683228;	P1[18][8] = 0.26301352;	P1[18][9] = 0.13780113;	P1[18][10] = -0.14467953;	P1[18][11] = 0.09699058;	P1[18][12] = -0.25476527;	P1[18][13] = -0.098759419;	P1[18][14] = -0.15325216;	P1[18][15] = 0.31649031;	P1[18][16] = -0.09805021;	P1[18][17] = -0.36061588;	P1[18][18] = -0.046404199;	P1[18][19] = 0.021186457;
		P1[19][0] = -0.010229702;	P1[19][1] = -0.071490404;	P1[19][2] = -0.33456422;	P1[19][3] = 0.19035667;	P1[19][4] = -0.15855578;	P1[19][5] = 0.13367485;	P1[19][6] = -0.10226951;	P1[19][7] = -0.037277677;	P1[19][8] = 0.16290187;	P1[19][9] = -0.046725004;	P1[19][10] = 0.069555789;	P1[19][11] = -0.39333354;	P1[19][12] = -0.28393251;	P1[19][13] = 0.10665249;	P1[19][14] = -0.07339062;	P1[19][15] = 0.05356669;	P1[19][16] = 0.70516439;	P1[19][17] = -0.00084960969;	P1[19][18] = 0.014824737;	P1[19][19] = 0.075915656;
		
		sum_letters[9] = 0.0866279;
		sum_letters[11] = 0.043972;
		sum_letters[5] = 0.0390894;
		sum_letters[7] = 0.0570451;
		sum_letters[18] = 0.0193078;
		sum_letters[14] = 0.0367281;
		sum_letters[4] = 0.0580589;
		sum_letters[10] = 0.0832518;
		sum_letters[16] = 0.0244313;
		sum_letters[3] = 0.048466;
		sum_letters[0] = 0.086209;
		sum_letters[2] = 0.0620286;
		sum_letters[17] = 0.0195027;
		sum_letters[12] = 0.0384319;
		sum_letters[13] = 0.0457631;
		sum_letters[1] = 0.0695179;
		sum_letters[6] = 0.0610127;
		sum_letters[19] = 0.0143859;
		sum_letters[15] = 0.0352742;
		sum_letters[8] = 0.0708956;

	}
	
	
	
	
	//######################################################################
	// Populate substitution matrix for the gap model
	// We will hardcode the matrix instead.
	//######################################################################
	
	lambda_gap[((int)('X' - 88)/-43)] = -0.002402885;
	lambda_gap[((int)('-' - 88)/-43)] = 0;
	
	P_gap[0][0] = -1/sqrt(2);
	P_gap[0][1] = 1/sqrt(2);
	P_gap[1][0] = 1/sqrt(2);
	P_gap[1][1] = 1/sqrt(2);
	
	P1_gap[0][0] = -1/sqrt(2);
	P1_gap[0][1] = 1/sqrt(2);
	P1_gap[1][0] = 1/sqrt(2);
	P1_gap[1][1] = 1/sqrt(2);
	
	double * sum_letters_gap;
	sum_letters_gap = new double[2];
	
	sum_letters_gap[((int)('X' - 88)/-43)] = 0.98219083;
	sum_letters_gap[((int)('-' - 88)/-43)] = 0.01780917;

	//######################################################################
	// Populate gamma rate parameters
	//######################################################################
	double * rateK;
	rateK = new double[nCatG];
	
	for(int i = 0;i<nCatG;++i){
		rateK[i] = ratei((i+1),nCatG,alpha);
	}
	
	//##########################################################################
	// We prescan the sequence to aggregate the gaps.
	//##########################################################################
	
	string prev_seg = "";
	int current_length = 0;
	int scalar_prev_segs = 0;
	double previous_dist;
	
	string prev_segs[gen.getSequenceByID(0).getSequence().length()];
	int * curr_lengths;
	
	curr_lengths = new int[gen.getSequenceByID(0).getSequence().length()];
	memset(curr_lengths,'\0',sizeof(int)*(gen.getSequenceByID(0).getSequence().length()));
	double * distance_average;	
	distance_average = new double[gen.getSequenceByID(0).getSequence().length()];
	memset(distance_average,'\0',sizeof(int)*(gen.getSequenceByID(0).getSequence().length()));
	double * power_law;
	power_law = new double[gen.getSequenceByID(0).getSequence().length()];

	for(int i = 0;i < gen.getSequenceByID(0).getSequence().length();++i){
		string col = "";
		for(int j = 0;j < parent;++j){
			if(nodes.getNodeByID(j).isParent() == 0){
				if(gen.getSequenceByID(j).getSequence().substr(i,1) == "-"){
					col += "-";
				}
				else{
					col += "X";
				}
			}
		}
		if(col == prev_seg){
			++current_length;
			previous_dist += distance_dist[i];
		}
		else{
			if(prev_seg != ""){
				prev_segs[scalar_prev_segs] = prev_seg;
				curr_lengths[scalar_prev_segs] = current_length;
				distance_average[scalar_prev_segs] = (previous_dist/((double)(current_length)));
				power_law[scalar_prev_segs] = pow(current_length,-1.5)/2.612;
				++scalar_prev_segs;
			}
			prev_seg = col;
			current_length = 1;
			previous_dist = distance_dist[i];
		}
		
	}
	
	prev_segs[scalar_prev_segs] = prev_seg;
	curr_lengths[scalar_prev_segs] = current_length;
	power_law[scalar_prev_segs] = pow(current_length,-1.5)/2.612;
	distance_average[scalar_prev_segs] = (previous_dist/((double)(current_length)));
	
	
	double * Pblas;
	Pblas = new double[400];
	double * Pblas2;
	Pblas2 = new double[400];

	double * TPblas;
	TPblas = new double[400];
	double * TPblas2;
	TPblas2 = new double[400];

	double * P2blas;
	P2blas = new double[400];
	double * P2blas2;
	P2blas2 = new double[400];
	double * gap_Pblas;
	gap_Pblas = new double[4];
	for(int it = 0;it < aalist_size_gap;++it){
		memcpy(&gap_Pblas[it*2],P1_gap[it],sizeof(double)*2);
	}
	for(int it = 0;it < aalist_size;++it){
		memcpy(&TPblas[it*20],P1[it],sizeof(double)*20);
	}


	//##########################################################################
	// GAP Process
	//##########################################################################
	
	int total_curr_length = 0;
	double * S_gap;
	
	double * P_gap_pos;

	double ** save_gap;
	double ** save2_gap;
	save_gap= new double * [2];
	save2_gap = new double * [2];
	
	for(int j = 0;j < 2;++j){
		save_gap[j] = new double [2];
		save2_gap[j] = new double [2];

		memset(save_gap[j],'\0',sizeof(double)*2);
		memset(save2_gap[j],'\0',sizeof(double)*2);
	}
	
	// This is the array of probabilities of the ancestral sequence.
	// P_gap[0][i] = gaps, P_gap[1][i] = amino acids.
	
	P_gap_pos = new double[gen.getSequenceByID(0).getSequence().length()]; 
	int aa_index;
	int aa_index2;
	int aa_index3;
	int gap_index_val;
	int gap_index_val2;
	int gap_index_val3;

	memset(P_gap_pos,'\0',sizeof(double)*(gen.getSequenceByID(0).getSequence().length()));
	double lambs_1_gap[aalist_size_gap];
	double lambs_2_gap[aalist_size_gap];
	for(int i = 0;i <= scalar_prev_segs;++i){
		S_gap = new double[parent * 2];
		memset(S_gap,'\0',sizeof(double)*(parent*2));
		
		string extant = "";	
		for(int j = 0;j < children_size;++j){
			string tmp = prev_segs[i];
			S_gap[j+ parent * ((int)(tmp[j] - 88)/-43)] = 1;
			extant += tmp[j];
		}	
		
		// Loop through parents;
		string letters;
		string letters2;
		string letters3;
		double P_1;
		double P_2;
		int child_1;
		int child_2;
		double distance_1;
		double distance_2;
		
		double col_gap = 0;
		for(int kcat = 0;kcat < nCatG;++kcat){
			double lgap = 0;
			for(int j = children_size;j < parent;++j){
				// Get children ID
				child_1 = nodes.getNodeByID(j).getChild_1();
				child_2 = nodes.getNodeByID(j).getChild_2();
				distance_1 = nodes.getNodeByID(j).getDistance_1() * distance_average[i] * rateK[kcat];
				distance_2 = nodes.getNodeByID(j).getDistance_2() * distance_average[i] * rateK[kcat];
				
				for(int it = 0;it < aalist_size_gap;++it){
					gap_index_val = (int)(aalist_gap[it][0] - 88)/-43;
					lambs_1_gap[it] = exp(lambda_gap[gap_index_val] * distance_1);
					lambs_2_gap[it] = exp(lambda_gap[gap_index_val] * distance_2);

					// First, copy each row to Pblas
					memcpy(&Pblas[it*2],P_gap[it],sizeof(double)*2);
					memcpy(&Pblas2[it*2],P_gap[it],sizeof(double)*2);
				}

				// Scale each row of P with lambs	
				for(int it = 0;it < aalist_size_gap;++it){			
					dscal(2,lambs_1_gap[it],&Pblas[it],2);
					dscal(2,lambs_2_gap[it],&Pblas2[it],2);
				}
				dgemm('T','T',2,2,2,1.0,Pblas,2,gap_Pblas,2,0,P2blas,2);
				dgemm('T','T',2,2,2,1.0,Pblas2,2,gap_Pblas,2,0,P2blas2,2);

				// Copy data to save
				for(int it = 0;it < aalist_size_gap;++it){
					memcpy(save_gap[it],&P2blas[it*2],sizeof(double)*2);
					memcpy(save2_gap[it],&P2blas2[it*2],sizeof(double)*2);
				}
				
				
				// Calculate likelihood
				for(int k = 0;k < aalist_size_gap;++k){
					gap_index_val = (int)(aalist_gap[k][0] - 88)/-43;
					// Calculate transition probabilities
					P_2 = 0;
					P_1 = 0;
					for(int n = 0;n < aalist_size_gap;++n){
						double p_temp1 = save_gap[k][n];
						double p_temp2 = save2_gap[k][n];
						
						P_1 += p_temp1 * S_gap[child_1 + parent * n];
						P_2 += p_temp2 * S_gap[child_2 + parent * n];
					}
					
					S_gap[j + parent * k] = (P_1 * P_2);
				}	
			}

			for(int k = 0;k < aalist_size_gap;++k){
				lgap += S_gap[(nodes.getNodeListSize()-1) + parent*k] * sum_letters_gap[k];
			}

			lgap *= power_law[i];
			// Gap likelihood must be distributed such that P(data|tree,substitution_model) * P(data|tree,gap_model) = likelihood
			// Data is divided in blocks of lengths k such that sum of all k = length of data
			// P(data|tree,gap_model) = P(block_1|tree,gap_model) * P(block_2|tree,gap_model) * ... * P(block_n|tree,gap_model)
			// To divide a block equally to each column k, the probabilities are preserved by : P(block_1|tree,gap_model) ^ (1/k) 

			col_gap += lgap/nCatG;
		}

		for(int k = 0;k < curr_lengths[i];++k){
			int pos = total_curr_length + k;
			P_gap_pos[pos] = log(col_gap)/curr_lengths[i];
		}
		
		delete [] S_gap;
		total_curr_length += curr_lengths[i];
	}
	//##########################################################################
	// Substitution process
	//#############################################################################
	double protein_likelihood;
	// To accelerate the process, we will add the observed columns in memory. 
	

	double ** save;
	double ** save2;
	save= new double * [20];
	save2 = new double * [20];
	
	for(int j = 0;j < 20;++j){
		save[j] = new double [20];
		save2[j] = new double [20];

		memset(save[j],'\0',sizeof(double)*20);
		memset(save2[j],'\0',sizeof(double)*20);
	}
	
	double * S;
	double lambs_1[aalist_size];
	double lambs_2[aalist_size];
	S = new double[parent * 20];
	bool valid_column = 0;
	for(int i = start -1;i < stop;++i){
		
		cout << i+1 << "	";
		memset(S,'\0',sizeof(double)*parent*20);
						
		string extant = "";
		valid_column = 0;
		for(int j = 0;j < children_size;++j){
				
			string tmp = nodes.getNodeByID(j).getSequence();
			
			if(tmp[i] == '-' || tmp[i] == 'X'){

			}
			else{
				S[j*20 + letter_order[(tmp[i] - 'A')]] = 1;
				valid_column = 1;
			}		
			extant += tmp[i];	
		}
		
		cout << extant;
		if(valid_column != 1){
			continue;
		}
		// Loop through parents;
		string letters;
		string letters2;
		string letters3;
		double P_1;
		double P_2;
		int child_1;
		int child_2;
		double distance_1;
		double distance_2;
		double col_likelihood = 0;
		double final_likelihood = 0;
		string max_letter = "";
		for(int kcat = 0;kcat < nCatG;++kcat){	
			for(int j = children_size;j < parent;++j){
				// Get children ID
				child_1 = nodes.getNodeByID(j).getChild_1();
				child_2 = nodes.getNodeByID(j).getChild_2();
				distance_1 = nodes.getNodeByID(j).getDistance_1() * distance_dist[i] * rateK[kcat];
				distance_2 = nodes.getNodeByID(j).getDistance_2() * distance_dist[i] * rateK[kcat];

				for(int it = 0;it < aalist_size;++it){
					aa_index = aalist[it][0] - 'A';
					lambs_1[it] = exp(lambda[aa_index] * distance_1/2);
					lambs_2[it] = exp(lambda[aa_index] * distance_2/2);

					// First, copy each row to Pblas
					memcpy(&Pblas[it*20],P[it],sizeof(double)*20);
					memcpy(&Pblas2[it*20],P[it],sizeof(double)*20);
				}
				
				// Scale each row of P with lambs	
				for(int it = 0;it < aalist_size;++it){			
					dscal(20,lambs_1[it],&Pblas[it],20);
					dscal(20,lambs_2[it],&Pblas2[it],20);
				}
				//dgemm('T','T',20,20,20,1.0,Pblas,20,TPblas,20,0,P2blas,20);
				//dgemm('T','T',20,20,20,1.0,Pblas2,20,TPblas,20,0,P2blas2,20);

				// Perform multiplication (A X^1/2) * (A X^1/2)**T
				dgemm('T','N',20,20,20,1.0,Pblas,20,Pblas,20,0,P2blas,20);
				dgemm('T','N',20,20,20,1.0,Pblas2,20,Pblas2,20,0,P2blas2,20);

				// Copy data to save
				for(int it = 0;it < aalist_size;++it){
					memcpy(save[it],&P2blas[it*20],sizeof(double)*20);
					memcpy(save2[it],&P2blas2[it*20],sizeof(double)*20);
				}
				
				
				
				// Calculate likelihood
				for(int k = 0;k < aalist_size;++k){
					aa_index = aalist[k][0] - 'A';
	
					// Calculate transition probabilities
					P_2 = 0;
					P_1 = 0;
					for(int n = 0;n < aalist_size;++n){
						double p_temp1 = 0;
						double p_temp2 = 0;
							
						p_temp1 = save[k][n];
						p_temp2 = save2[k][n];			
																	
						P_1 += p_temp1 * S[child_1 * 20 + n];
						P_2 += p_temp2 * S[child_2 * 20 + n];
					}
						
					if(P_2 == 0){
						P_2 = 1;	
					}
					if(P_1 == 0){
						P_1 = 1;	
					}
					if(P_1 == 1 && P_2 == 1){
						P_1 = 0;
						P_2 = 0;
					}
					// Above is deal with missing data	
					S[j * 20 + k] = (P_1 * P_2);
				}
				
			}
			// Compute final likelihood
			// Compute maximal likelihood
			double likelihood = 0;
			double max_likelihood = 0;
			
			double L = 0;
					
			for(int k = 0;k < aalist_size;++k){
				L = S[(nodes.getNodeListSize()-1) * 20 + k] * sum_letters[k];
				likelihood += L;
			}
			col_likelihood += likelihood/nCatG;
		}
		final_likelihood = log(col_likelihood);
		final_likelihood += P_gap_pos[i];
		
		string tosave;
		stringstream out;
		out.precision(20);
		out << fixed << final_likelihood;
		tosave = "	" + max_letter + "	" + out.str();
		out.clear();
		protein_likelihood += final_likelihood;
		cout << tosave <<endl;
		
		

	}
	delete [] S;
	cout << endl;
	cout << endl;
	cout <<  "===================================================" << endl;
	cout <<  "Analyzed " << fastafile << endl;;
	cout <<  "Log likelihood of the protein under tree is : " << protein_likelihood << endl;
	
	cout <<  "Tree used : " << nodes.print_tree(parent-1) << endl;
	cout <<  "===================================================" << endl;
	delete[] curr_lengths;
	delete[] distance_average;
	delete[] lambda;
	delete[] sum_letters;
	delete[] sum_letters_gap;
	delete[] P_gap_pos;
	
	return 0;
	
}
